#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-08-24
# @Filename: notifier.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import hashlib
import logging
from time import time
from traceback import format_exception

from typing import TYPE_CHECKING, Any, Literal, Protocol, cast

import httpx

from sdsstools import Configuration

from gort import config
from gort.tools import LogNamespace


if TYPE_CHECKING:
    from gort.gort import Gort


NotificationLevel = Literal["debug", "info", "warning", "error", "critical"]

GORT_ICON_URL = "https://github.com/sdss/lvmgort/blob/main/docs/sphinx/_static/gort_logo_slack.png?raw=true"


class OverwatcherProtocol(Protocol):
    gort: Gort
    log: LogNamespace
    config: Configuration


class NotifierMixIn(OverwatcherProtocol):
    """A mix-in class for :obj:`.Overwatcher `that adds notification capabilities."""

    # A dictionary of notification hash and the timestamp
    # when the notification can be sent again.
    notification_history: dict[str, float] = {}

    async def notify(
        self,
        message: str | None = None,
        level: NotificationLevel | None = None,
        error: str | Exception | None = None,
        with_traceback: bool = True,
        slack_channel: str | bool | None = None,
        database: bool = True,
        log: bool = True,
        payload: dict[str, Any] = {},
        as_overwatcher: bool = True,
        allow_repeat_notifications: bool = False,
        min_time_between_repeat_notifications: int = 60,
    ):
        """Emits a notification to Slack or email.

        The notification is logged to the GORT/Overwatcher log and to the
        database. Depending on severity and arguments, a notification is then
        issued over Slack to the appropriate channel, or over email for serious
        alerts.

        Parameters
        ----------
        message
            The message to send.
        level
            The level of the message. One of 'debug', 'info', 'warning', 'error',
            or 'critical'. If :obj:`None`, the level is set to ``error`` if
            ``error`` is provided, and to ``info`` otherwise. Critical errors
            are sent to the ``lvm-alerts`` Slack channel.
        error
            An error message or exception to include in the notification.
        with_traceback
            Whether to include the traceback in the notification. Requires
            ``error`` to be an exception object.
        slack_channel
            The Slack channel to which to send the notification. By default
            ``lvm-alerts`` is notified for ``critical`` messages,
            and ``lvm-overwatcher`` for anything lower. If ``False``, no
            Slack notifications are sent.
        database
            Whether to record the notification in the database.
        log
            Whether to record the notification in the log.
        payload
            Additional notification payload as a JSON-like dictionary. Only
            saved to the database notifications table.
        as_overwatcher
            Whether to send the message as the Overwatcher bot.
        allow_repeat_notifications
            Whether to allow the same notification to be sent multiple times.
        min_time_between_repeat_notifications
            The minimum time in seconds between repeated notifications. Ignored
            if ``allow_repeat_notifications`` is ``False``. This only affects Slack
            and email notifications. The notification is always recorded to the
            database and log if those options are passed.

        """

        if level is None:
            level = "error" if error is not None else "info"

        message = message or ""
        if error is not None and message == "":
            message = f" {str(error)}"

        trace: str | None = None
        if with_traceback and isinstance(error, Exception):
            trace = "".join(format_exception(type(error), error, error.__traceback__))

        full_message = message
        if trace:
            full_message += f"\n{trace}" if full_message else trace

        if log:
            log_level = logging._nameToLevel[level.upper()]
            self.log.logger.log(log_level, self.log._get_message(full_message))

        # Now create the notification actual notification by calling the API.
        # This will load it to the database. We do not emit emails for now.
        api_host, api_port = config["services"]["lvmapi"].values()

        slack_config = self.config["overwatcher.slack"]
        if slack_channel is None or slack_channel is True:
            slack_channel = cast(str, slack_config["notifications_channel"])

        # Create a notification hash to uniquely identify the notification.
        notification_hash = self.create_notification_hash(
            message=message,
            level=level,
            error=error,
            slack_channel=slack_channel,
            payload=payload,
        )

        # Do not emit Slack notification if this is a repeated notification
        next_notification_time = self.notification_history.get(notification_hash)
        if (
            not allow_repeat_notifications
            and next_notification_time
            and next_notification_time > time()
        ):
            slack_channel = False

        async with httpx.AsyncClient(
            base_url=f"http://{api_host}:{api_port}",
            follow_redirects=True,
        ) as client:
            response = await client.post(
                "/notifications/create",
                json={
                    "message": full_message,
                    "level": level.upper(),
                    "payload": payload,
                    "slack_channel": slack_channel,
                    "email_on_critical": False,
                    "write_to_database": database,
                    "slack_extra_params": {
                        "username": "Overwatcher" if as_overwatcher else None,
                        "icon_url": GORT_ICON_URL if as_overwatcher else None,
                    },
                },
            )

            code = response.status_code
            if code != 200:
                self.log.warning(f"Failed adding night log comment. Code {code}.")

        next_notification_time = time() + min_time_between_repeat_notifications
        self.notification_history[notification_hash] = next_notification_time

    def create_notification_hash(
        self,
        message: str | None = None,
        level: NotificationLevel | None = None,
        error: str | Exception | None = None,
        slack_channel: str | bool | None = None,
        payload: dict[str, Any] = {},
    ):
        """Creates a hash for a notification."""

        hash_elements = [
            str(message) if message else "",
            level or "",
            str(error) if error else "",
            str(slack_channel) if isinstance(slack_channel, str) else "",
            str(payload) if payload else "",
        ]

        hasher = hashlib.new("md5")
        hasher.update("".join(hash_elements).encode())

        return hasher.hexdigest()


class BasicNotifier(NotifierMixIn):
    """Basic notifier."""

    def __init__(self, gort: Gort):
        self.gort = gort
        self.config = Configuration(gort.config)
        self.log = LogNamespace(gort.log)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-08-13
# @Filename: base.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Type

from gort.enums import Event
from gort.pubsub import notify_event


if TYPE_CHECKING:
    from gort.gort import Gort


__all__ = ["recipes", "BaseRecipe"]


recipes: dict[str, Type[BaseRecipe]] = {}


class RegisterRecipe(type):
    """Metaclass to register recipes."""

    def __new__(cls, name, bases, class_dict):
        cls = type.__new__(cls, name, bases, class_dict)

        if name == "BaseRecipe":
            return cls

        assert issubclass(cls, BaseRecipe)

        if "name" not in class_dict:
            raise ValueError(f"name attribute not defined in {name}.")

        # This is a hacked version of abc.ABCMeta, but this way we
        # don't need to chain two metaclasses.
        if "recipe" not in class_dict:
            raise ValueError(f"recipe method not defined in {name}.")

        recipes[class_dict["name"]] = cls

        return cls


class BaseRecipe(object, metaclass=RegisterRecipe):
    """Base class for recipes."""

    name: str | None = None

    def __init__(self, gort: Gort):
        self.gort = gort

    async def recipe(self, *args, **kwargs):
        """The recipe. Must be overridden by the subclass."""

        return

    async def __call__(self, *args, **kwargs):
        """Executes the recipe and sends event notifications."""

        payload: dict[str, Any] = {
            "recipe_name": self.name,
            "args": list(args),
            "kwargs": kwargs,
        }

        await notify_event(Event.RECIPE_START, payload=payload)

        try:
            await self.recipe(*args, **kwargs)
        except Exception as err:
            error_payload = payload.copy()
            error_payload["error"] = str(err)
            await notify_event(Event.RECIPE_FAILED, payload=error_payload)
            raise
        else:
            await notify_event(Event.RECIPE_END, payload=payload)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-11-05
# @Filename: recipes.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import abc
import asyncio

from typing import TYPE_CHECKING, ClassVar

from gort.enums import ErrorCode
from gort.exceptions import TroubleshooterCriticalError
from gort.tools import set_tile_status


if TYPE_CHECKING:
    from .troubleshooter import TroubleModel, Troubleshooter


__all__ = ["TroubleshooterRecipe"]


class TroubleshooterRecipe(metaclass=abc.ABCMeta):
    """Base class for troubleshooter recipes."""

    priority: ClassVar[int | None] = None
    name: ClassVar[str]

    def __init__(self, troubleshooter: Troubleshooter):
        self.troubleshooter = troubleshooter

        self.overwatcher = troubleshooter.overwatcher
        self.gort = troubleshooter.overwatcher.gort

        self.notify = troubleshooter.overwatcher.notify

        assert hasattr(self, "name"), "Recipe must have a name."

    @abc.abstractmethod
    def match(self, error_model: TroubleModel) -> bool:
        """Returns True if the recipe can handle the error."""

        raise NotImplementedError

    async def handle(self, error_model: TroubleModel) -> bool:
        """Runs the recipe to handle the error."""

        try:
            return await self._handle_internal(error_model)
        except TroubleshooterCriticalError:
            # Propagate this error, which will be handled by the Troubleshooter class.
            raise
        except Exception as err:
            await self.notify(
                f"Error running recipe {self.name}: {err!r}",
                level="error",
            )
            return False

    @abc.abstractmethod
    async def _handle_internal(self, error_model: TroubleModel) -> bool:
        """Internal implementation of the recipe."""

        raise NotImplementedError


class CleanupRecipe(TroubleshooterRecipe):
    """A recipe that cleans up after an error."""

    priority = 500
    name = "cleanup"

    def match(self, error_model: TroubleModel) -> bool:
        """Returns True if the recipe can handle the error."""

        # Always return False. This is a last resort recipe.
        return False

    async def _handle_internal(self, error_model: TroubleModel) -> bool:
        """Run the cleanup recipe."""

        await self.overwatcher.gort.cleanup(readout=False)

        return True


class AcquisitionFailedRecipe(TroubleshooterRecipe):
    """Handle acquisition failures."""

    priority = 1
    name = "acquisition_failed"

    def match(self, error_model: TroubleModel) -> bool:
        """Returns True if the recipe can handle the error."""

        if error_model.error_code == ErrorCode.ACQUISITION_FAILED:
            return True

        return False

    async def _handle_internal(self, error_model: TroubleModel) -> bool:
        """Handle the error."""

        error = error_model.error

        tile_id: int | None = error.payload.get("tile_id", None)
        if tile_id is None:
            await self.notify(
                'Cannot disable tile without a "tile_id. '
                "Continuing observations without disabling tile.",
                level="error",
            )
        else:
            await set_tile_status(tile_id, enabled=False)
            await self.notify(
                f"tile_id={tile_id} has been disabled. Continuing observations.",
                level="warning",
            )

        return True


class SchedulerFailedRecipe(TroubleshooterRecipe):
    """Handle acquisition failures."""

    priority = 1
    name = "scheduler_failed"

    def match(self, error_model: TroubleModel) -> bool:
        """Returns True if the recipe can handle the error."""

        if error_model.error_code.is_scheduler_error():
            return True

        return False

    async def _handle_internal(self, error_model: TroubleModel) -> bool:
        """Handle the error."""

        await self.notify(
            "The scheduler was not able to find a valid tile to "
            "observe. Waiting 60 seconds before trying again.",
            level="warning",
        )
        await asyncio.sleep(60)

        return True


class SpectrographNotIdleRecipe(TroubleshooterRecipe):
    """Handle acquisition failures."""

    priority = 1
    name = "spectrograph_not_idle"

    def match(self, error_model: TroubleModel) -> bool:
        """Returns True if the recipe can handle the error."""

        if error_model.error_code == ErrorCode.SECTROGRAPH_NOT_IDLE:
            return True

        return False

    async def _handle_internal(self, error_model: TroubleModel) -> bool:
        """Handle the error."""

        await self.notify("Resetting spectrographs.", level="warning")
        await self.gort.specs.reset()

        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-10-27
# @Filename: dome.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import enum

from typing import TYPE_CHECKING

from lvmopstools.retrier import Retrier

from sdsstools.utils import GatheringTaskGroup

from gort.exceptions import GortError


if TYPE_CHECKING:
    from gort.overwatcher.overwatcher import Overwatcher


__all__ = ["DomeHelper", "DomeStatus"]


class DomeStatus(enum.Flag):
    """An enumeration of dome statuses."""

    OPEN = enum.auto()
    CLOSED = enum.auto()
    OPENING = enum.auto()
    CLOSING = enum.auto()
    MOVING = enum.auto()
    UNKNOWN = enum.auto()


class DomeHelper:
    """Handle dome movement."""

    def __init__(self, overwatcher: Overwatcher):
        self.overwatcher = overwatcher
        self.gort = overwatcher.gort
        self.log = overwatcher.log

        self._action_lock = asyncio.Lock()
        self._move_lock = asyncio.Lock()

    @Retrier(max_attempts=3, delay=1)
    async def status(self):
        """Returns the status of the dome."""

        status = await self.gort.enclosure.status()
        labels = status["dome_status_labels"]

        if "MOTOR_OPENING" in labels:
            return DomeStatus.OPENING | DomeStatus.MOVING
        elif "MOTOR_CLOSING" in labels:
            return DomeStatus.CLOSING | DomeStatus.MOVING
        elif "OPEN" in labels:
            if "MOVING" in labels:
                return DomeStatus.OPENING | DomeStatus.MOVING
            return DomeStatus.OPEN
        elif "CLOSED" in labels:
            if "MOVING" in labels:
                return DomeStatus.CLOSING | DomeStatus.MOVING
            return DomeStatus.CLOSED

        return DomeStatus.UNKNOWN

    async def is_opening(self):
        """Returns True if the dome is open or opening."""

        status = await self.status()

        if status & DomeStatus.OPENING:
            return True
        if status & DomeStatus.OPEN:
            return True

        return False

    async def is_closing(self):
        """Returns True if the dome is closed or closed."""

        status = await self.status()

        if status & DomeStatus.CLOSING:
            return True
        if status & DomeStatus.CLOSED:
            return True

        return False

    async def is_moving(self):
        """Returns True if the dome is moving."""

        status = await self.status()

        if status & DomeStatus.MOVING:
            return True

        return False

    async def wait_until_idle(self):
        """Waits until the dome is idle."""

        while await self.is_moving():
            await asyncio.sleep(5)

    @Retrier(max_attempts=2, delay=5)
    async def _move_with_retries(
        self,
        open: bool = False,
        park: bool = True,
        retry_without_parking: bool = False,
    ):
        """Moves the dome with retries."""

        if open:
            await self.gort.enclosure.open(park_telescopes=park)
        else:
            await self.gort.enclosure.close(
                park_telescopes=park,
                retry_without_parking=retry_without_parking,
                force=True,
            )

    async def _move(
        self,
        status: DomeStatus,
        open: bool = False,
        park: bool = True,
        retry_without_parking: bool = False,
    ):
        """Moves the dome."""

        if open and not self.overwatcher.state.safe:
            raise GortError("Cannot open the dome when conditions are unsafe.")

        is_local = await self.gort.enclosure.is_local()
        if is_local:
            raise GortError("Cannot move the dome in local mode.")

        failed: bool = False

        try:
            if status & DomeStatus.MOVING:
                self.log.warning("Dome is already moving. Stopping.")
                await self.stop()

            await self._move_with_retries(
                open=open,
                park=park,
                retry_without_parking=retry_without_parking,
            )

        except Exception:
            await asyncio.sleep(3)

            status = await self.status()

            if status & DomeStatus.MOVING:
                self.log.warning("Dome is still moving after an error. Stopping.")
                await self.stop()

            # Sometimes the open/close could fail but actually the dome is open/closed.
            if open and not (status & DomeStatus.OPEN):
                failed = True
            elif not open and not (status & DomeStatus.CLOSED):
                failed = True

            if failed:
                await self.overwatcher.notify(
                    "The dome has failed to open/close. Disabling the Overwatcher "
                    "to prevent further attempts. Please check the dome immediately, "
                    "it may be partially or fully open.",
                    level="critical",
                )

                # Release the lock here. force_disable() may require closing the dome.
                if self._move_lock.locked():
                    self._move_lock.release()

                await self.overwatcher.force_disable()
                raise

    async def open(self, park: bool = True):
        """Opens the dome."""

        async with self._move_lock:
            status = await self.status()
            if status == DomeStatus.OPEN:
                self.log.debug("Dome is already open.")
                return

            if status == DomeStatus.OPENING:
                self.log.debug("Dome is already opening.")
                return

            if status == DomeStatus.UNKNOWN:
                self.log.warning("Dome is in an unknown status. Stopping and opening.")
                await self.stop()
                status = await self.status()

            await self._move(status, open=True, park=park)

    async def close(self, park: bool = True, retry: bool = False):
        """Closes the dome."""

        async with self._move_lock:
            status = await self.status()
            if status == DomeStatus.CLOSED:
                self.log.debug("Dome is already closed.")
                return

            if status == DomeStatus.CLOSING:
                self.log.debug("Dome is already closing.")
                return

            if status == DomeStatus.UNKNOWN:
                self.log.warning("Dome is in an unknown status. Stopping and closing.")
                await self.stop()
                status = await self.status()

            await self._move(
                status,
                open=False,
                park=park,
                retry_without_parking=retry,
            )

    async def stop(self):
        """Stops the dome."""

        await self.gort.enclosure.stop()

        await asyncio.sleep(1)
        status = await self.status()

        if status & DomeStatus.MOVING:
            raise GortError("Dome is still moving after a stop command.")

    async def startup(self):
        """Runs the startup sequence."""

        async with self._action_lock:
            self.log.info("Starting the dome startup sequence.")
            await self.gort.startup(open_enclosure=False, focus=False)

            # Now we manually open. We do not focus here since that's handled
            # by the observer module.
            await self.open()

    async def shutdown(
        self,
        retry: bool = False,
        force: bool = False,
        park: bool = True,
    ):
        """Runs the shutdown sequence."""

        is_closing = await self.is_closing()
        if is_closing and not force:
            return

        async with self._action_lock:
            self.log.info("Running the shutdown sequence.")

            async with GatheringTaskGroup() as group:
                self.log.info("Turning off all lamps.")
                group.create_task(self.gort.nps.calib.all_off())

                self.log.info("Making sure guiders are idle.")
                group.create_task(self.gort.guiders.stop())

                self.log.info("Closing the dome.")
                group.create_task(self.close(retry=retry))

            self.log.info("Parking telescopes for the night.")
            if park:
                await self.gort.telescopes.park()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-03-10
# @Filename: telescope.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from collections import defaultdict
from time import time

from typing import TYPE_CHECKING, ClassVar

import numpy

from gort.devices.core import GortDevice, GortDeviceSet
from gort.enums import Event
from gort.exceptions import ErrorCode, GortTelescopeError
from gort.tools import angular_separation, kubernetes_restart_deployment


if TYPE_CHECKING:
    from gort.gort import Gort
    from gort.remote import ActorReply


__all__ = ["Telescope", "TelescopeSet", "KMirror", "FibSel", "Focuser", "MoTanDevice"]


class MoTanDevice(GortDevice):
    """A TwiceAsNice device."""

    #: Artificial delay introduced to prevent all motors to slew at the same time.
    SLEW_DELAY: ClassVar[float | dict[str, float]] = 0

    device_type: ClassVar[str]

    def __init__(self, gort: Gort, name: str, actor: str):
        super().__init__(gort, name, actor)

        class_name = self.__class__.__name__

        timeouts = self.gort.config["telescopes"]["timeouts"].get(class_name.lower())

        self.timeouts: defaultdict[str, float | None]
        self.timeouts = defaultdict(lambda: None, timeouts)

        self.telescope = self.name.split(".")[0]

        self.device: str
        if "fibsel" in self.name:
            self.device = "fibsel"
        elif "foc" in self.name:
            self.device = "focuser"
        elif "km" in self.name:
            self.device = "kmirror"
        else:
            raise GortTelescopeError(f"Invalid device type {self.name!r}.")

    async def status(self):
        """Returns the status of the device."""

        return await self.run_command("status")

    async def is_reachable(self):
        """Is the device reachable?"""

        is_reachable = await self.run_command("isReachable")

        return bool(is_reachable.get("Reachable"))

    async def check_reachable(self):
        """Checks if the device is reacheable or issues an error."""

        if not (await self.is_reachable()):
            raise GortTelescopeError(
                f"MoTAN device {self.name!r} is not reachable.",
                error_code=ErrorCode.MOTAN_DEVICE_NOT_REACHABLE,
                payload={"telescope": self.telescope, "device": self.device},
            )

    async def is_moving(self):
        """Is the device moving."""

        is_moving = await self.run_command("isMoving")

        return bool(is_moving.get("Moving"))

    async def slew_delay(self):
        """Sleeps the :obj:`.SLEW_DELAY` amount."""

        if isinstance(self.SLEW_DELAY, (float, int)):
            await asyncio.sleep(self.SLEW_DELAY)
        else:
            await asyncio.sleep(self.SLEW_DELAY[self.telescope])

    async def stop(self):
        """Stop the K-mirror movement."""

        if await self.is_moving():
            self.write_to_log("Stopping slew.")

            if self.device_type == "kmirror":
                await self.run_command("slewStop", timeout=self.timeouts["slewStop"])
            await self.run_command("stop", timeout=self.timeouts["slewStop"])

            if await self.is_moving():
                raise GortTelescopeError(
                    f"Failed to stop the {self.name}.",
                    error_code=ErrorCode.MOTAN_DEVICE_CANNOT_STOP,
                    payload={"telescope": self.telescope, "device": self.device},
                )

    async def run_command(
        self,
        command: str,
        *args,
        n_retries: int = 3,
        delay: float = 1,
        **kwargs,
    ) -> ActorReply:
        """Runs a MoTan command with retries."""

        return await self.actor.commands[command](
            *args,
            n_retries=n_retries,
            delay=delay,
            **kwargs,
        )


class KMirror(MoTanDevice):
    """A device representing a K-mirror."""

    SLEW_DELAY = 0

    device_type = "kmirror"

    async def home(self):
        """Homes the k-mirror."""

        await self.check_reachable()

        await self.slew_delay()
        await self.stop()

        self.write_to_log("Homing k-mirror.", level="info")
        await self.run_command("moveToHome", timeout=self.timeouts["moveToHome"])
        self.write_to_log("k-mirror homing complete.")

    async def park(self):
        """Park the k-mirror at 90 degrees."""

        await self.move(90)

    async def move(self, degs: float):
        """Move the k-mirror to a position in degrees. Does NOT track after the move.

        Parameters
        ----------
        degs
            The position to which to move the k-mirror, in degrees.

        """

        await self.check_reachable()

        await self.slew_delay()
        await self.stop()

        self.write_to_log(f"Moving k-mirror to {degs:.3f} degrees.", level="info")
        await self.run_command(
            "moveAbsolute",
            degs,
            "deg",
            timeout=self.timeouts["moveAbsolute"],
        )

    async def slew(
        self,
        ra: float,
        dec: float,
        offset_angle: float = 0.0,
        stop_degs_before: float = 0.0,
    ):
        """Moves the mirror to the position for ``ra, dec`` and starts slewing.

        Parameters
        ----------
        ra
            Right ascension of the field to track, in degrees.
        dec
            Declination of the field to track, in degrees.
        offset_angle
            Derotation offset in degrees. This value is converted to the
            -180 to 180 deg range before sending it to the k-mirror.
        stop_degs_before
            Number of degrees to stop before reaching the desired position
            angle. This has the effect of actually slewing to
            ``offset_angle-stop_degs_before``. Useful if we want to be
            sure that positive offsets will be applied without backlash.

        """

        await self.check_reachable()

        await self.slew_delay()

        offset_angle %= 360
        if offset_angle > 180:
            offset_angle -= 360

        if offset_angle == 0:
            msg = f"Slewing k-mirror to ra={ra:.6f} dec={dec:.6f} and tracking."
        else:
            msg = (
                f"Slewing k-mirror to ra={ra:.6f} dec={dec:.6f} "
                f"pa={offset_angle:.3f} and tracking."
            )

        self.write_to_log(msg, level="info")

        stop_degs_before = abs(stop_degs_before)
        if abs(stop_degs_before) > 0:
            self.write_to_log(f"Using stop_degs_before={stop_degs_before}.")

        await self.stop()
        await self.run_command(
            "slewStart",
            ra / 15.0,
            dec,
            seg_time=self.gort.config["telescopes"]["kmirror"]["seg_time"],
            seg_min_num=self.gort.config["telescopes"]["kmirror"]["seg_min_num"],
            offset_angle=offset_angle - stop_degs_before,
            timeout=self.timeouts["slewStart"],
        )


class Focuser(MoTanDevice):
    """A device representing a focuser."""

    SLEW_DELAY = 0

    device_type = "focuser"

    async def home(self):
        """Homes the focuser.

        Parameters
        ----------
        restore_position
            Whether to restore the previous focuser position after homing.

        """

        await self.check_reachable()

        await self.slew_delay()
        await self.stop()

        # Store current position to restore it later.
        status = await self.status()
        current_position = status.get("Position")

        self.write_to_log("Homing focuser.", level="info")
        await self.run_command("moveToHome", timeout=self.timeouts["moveToHome"])
        self.write_to_log("Focuser homing complete.")

        if current_position is not None and not numpy.isnan(current_position):
            self.write_to_log(f"Restoring position {current_position} DT.")
            await self.move(current_position)

    async def move(self, dts: float):
        """Move the focuser to a position in DT."""

        await self.check_reachable()

        await self.slew_delay()
        await self.stop()

        self.write_to_log(f"Moving focuser to {dts:.3f} DT.", level="info")
        await self.run_command(
            "moveAbsolute",
            dts,
            "DT",
            timeout=self.timeouts["moveAbsolute"],
        )


class FibSel(MoTanDevice):
    """A device representing the fibre mask in the spectrophotometric telescope."""

    # We really don't want a delay here because it would slow down the acquisition
    # of new standards, and anyway the fibre selector usually moves by itself.
    SLEW_DELAY: float = 0

    # Rehome after this many seconds.
    HOME_AFTER: float | None = None

    device_type = "fibsel"

    def __init__(self, gort: Gort, name: str, actor: str):
        super().__init__(gort, name, actor)

        self.__last_homing: float = 0

    async def home(self):
        """Homes the fibre selector."""

        await self.check_reachable()

        await self.slew_delay()
        await self.stop()

        self.write_to_log("Homing fibsel.", level="info")
        await self.run_command("moveToHome", timeout=self.timeouts["moveToHome"])
        self.write_to_log("Fibsel homing complete.")

        self.__last_homing = time()

    def list_positions(self) -> list[str]:
        """Returns a list of valid positions."""

        return list(self.gort.config["telescopes"]["mask_positions"])

    async def _check_home(self):
        """Checks if a homing is required before moving the mask."""

        if self.HOME_AFTER is None:
            return

        if time() - self.__last_homing > self.HOME_AFTER:
            self.write_to_log("Rehoming fibsel before moving.", "warning")
            await self.home()

    async def move_to_position(self, position: str | int, rehome: bool = False):
        """Moves the spectrophotometric mask to the desired position.

        Parameters
        ----------
        position
            A position in the form `PN-M` where ``N=1,2`` and ``M=1-12``, in which
            case the mask will rotate to expose the fibre with that name. If
            ``position`` is a number, moves the mask to that value.
        rehome
            Home the fibre selector before moving to the new position.

        """

        await self.check_reachable()

        if rehome:
            await self.home()

        if isinstance(position, str):
            mask_positions = self.gort.config["telescopes"]["mask_positions"]
            if position not in mask_positions:
                raise GortTelescopeError(
                    f"Cannot find position {position!r}.",
                    error_code=ErrorCode.FIBSEL_INVALID_POSITION,
                )

            steps = mask_positions[position]
            self.write_to_log(f"Moving mask to {position}: {steps} DT.", level="info")

        else:
            steps = position
            self.write_to_log(f"Moving mask to {steps} DT.", level="info")

        await self._move("moveAbsolute", steps)

    async def move_relative(self, steps: float):
        """Move the mask a number of motor steps relative to the current position."""

        await self.check_reachable()

        self.write_to_log(f"Moving fibre mask {steps} steps.")
        await self._move("moveRelative", steps)

    async def _move(self, command: str, steps: float, allow_rehoming: bool = True):
        """Moves the fibre selector. If the move fails, tries rehoming."""

        try:
            await self._check_home()

            await self.slew_delay()
            await self.stop()

            await self.run_command(command, steps, timeout=self.timeouts[command])
        except Exception as err:
            if allow_rehoming:
                self.write_to_log(
                    f"Failed to move fibsel with error: {err} - Rehoming and retrying.",
                    "warning",
                )

                # Notify this event so that it can be taken into account, for
                # example to add a flag if this happens during the standards
                # iteration loop.
                await self.gort.notify_event(
                    Event.UNEXPECTED_FIBSEL_REHOME,
                    payload={"time": time()},
                )

                await self.home()
                await self._move(command, steps, allow_rehoming=False)
            else:
                raise


class Telescope(GortDevice):
    """Class representing an LVM telescope functionality."""

    def __init__(self, gort: Gort, name: str, actor: str, **kwargs):
        super().__init__(gort, name, actor)

        self.is_homed: bool = False

        self.config = self.gort.config["telescopes"]
        self.pwi = self.actor

        kmirror_actor = kwargs.get("kmirror", None)
        if kmirror_actor:
            self.has_kmirror = True
            self.km = KMirror(self.gort, f"{self.name}.km", kmirror_actor)
        else:
            self.has_kmirror = False
            self.km = None

        self.focuser = Focuser(self.gort, f"{self.name}.focuser", kwargs["focuser"])

        self.fibsel = (
            FibSel(self.gort, f"{self.name}.fibsel", "lvm.spec.fibsel")
            if self.name == "spec"
            else None
        )

        if self.name in self.gort.guiders:
            self.guider = self.gort.guiders[name]
        else:
            self.guider = None

        self.timeouts = self.gort.config["telescopes"]["timeouts"]["pwi"]

    async def init(self):
        """Determines the initial state of the telescope."""

        # If the axes are enabled, we assume the telescope is homed.
        if not self.is_homed and (await self.is_ready()):
            self.is_homed = True

    async def status(self):
        """Retrieves the status of the telescope."""

        reply: ActorReply = await self.pwi.commands.status()
        return reply.flatten()

    async def is_parked(self):
        """Is the telescope parked?"""

        park_position = self.gort.config["telescopes"]["named_positions"]["park"]["all"]

        status = await self.status()
        if status["is_enabled"] or status["is_tracking"] or status["is_slewing"]:
            return False

        alt_diff = numpy.abs(status["altitude_degs"] - park_position["alt"])
        az_diff = numpy.abs(status["azimuth_degs"] - park_position["az"])

        if alt_diff > 5 or az_diff > 5:
            return False

        return True

    async def is_ready(self):
        """Checks if the telescope is ready to be moved."""

        status = await self.status()

        is_connected = status.get("is_connected", False)
        is_enabled = status.get("is_enabled", False)

        return is_connected and is_enabled

    async def initialise(self, home: bool | None = None):
        """Connects to the telescope and initialises the axes.

        Parameters
        ----------
        home
            If :obj:`True`, runs the homing routine after initialising.

        """

        if not (await self.is_ready()):
            self.write_to_log("Initialising telescope.")
            await self.pwi.commands.setConnected(True)
            await self.pwi.commands.setEnabled(True)

        if home is True:
            await self.home()

    async def home(
        self,
        home_telescope: bool = True,
        home_km: bool = True,
        home_focuser: bool = False,
        home_fibsel: bool = False,
    ):
        """Initialises and homes the telescope.

        Parameters
        ---------
        home_telescope
            Homes the telescope. Defaults to :obj:`True`.
        home_km
            Homes the K-mirror, if present. Defaults to :obj:`True`.
        home_focuser
            Homes the focuser. Defaults to :obj:`False`.
        home_fibsel
            Homes the fibre selector, if present. Defaults to :obj:`False`.

        """

        home_subdevices_task: asyncio.Future | None = None

        subdev_tasks = []
        if self.km is not None and home_km:
            subdev_tasks.append(self.km.home())
        if self.fibsel is not None and home_fibsel:
            subdev_tasks.append(self.fibsel.home())
        if self.focuser is not None and home_focuser:
            subdev_tasks.append(self.focuser.home())

        home_subdevices_task = asyncio.gather(*subdev_tasks)

        if home_telescope:
            if await self.gort.enclosure.is_local():
                raise GortTelescopeError(
                    "Cannot home in local mode.",
                    error_code=ErrorCode.CANNOT_MOVE_LOCAL_MODE,
                )

            self.write_to_log("Homing telescope.", level="info")

            if not (await self.is_ready()):
                await self.pwi.commands.setConnected(True)
                await self.pwi.commands.setEnabled(True)

            await self.pwi.commands.findHome()

            # findHome does not block, so wait a reasonable amount of time.
            await asyncio.sleep(15)

            self.is_homed = True

        if home_subdevices_task is not None and not home_subdevices_task.done():
            await home_subdevices_task

    async def park(
        self,
        disable=True,
        use_pw_park=False,
        alt_az: tuple[float, float] | None = None,
        kmirror: bool = True,
        force: bool = False,
    ):
        """Parks the telescope.

        Parameters
        ----------
        disable
            Disables the axes after reaching the park position.
        use_pw_park
            Uses the internal park position stored in the PlaneWave mount.
        alt_az
            A tuple with the alt and az position at which to park the telescope.
            If not provided, defaults to the ``park`` named position.
        kmirror
            Whether to park the mirror at 90 degrees.
        force
            Moves the telescope even if the mode is local.

        """

        if await self.gort.enclosure.is_local():
            raise GortTelescopeError(
                "Cannot home in local mode.",
                error_code=ErrorCode.CANNOT_MOVE_LOCAL_MODE,
            )

        await self.initialise()

        if use_pw_park:
            self.write_to_log("Parking telescope to PW default position.", level="info")
            await self.pwi.commands.park()
        elif alt_az is not None:
            self.write_to_log(f"Parking telescope to alt={alt_az[0]}, az={alt_az[1]}.")
            await self.goto_coordinates(
                alt=alt_az[0],
                az=alt_az[1],
                kmirror=False,
                altaz_tracking=False,
                force=force,
            )
        else:
            coords = self.gort.config["telescopes"]["named_positions"]["park"]["all"]
            alt = coords["alt"]
            az = coords["az"]
            self.write_to_log(f"Parking telescope to alt={alt}, az={az}.", level="info")
            await self.goto_coordinates(
                alt=alt,
                az=az,
                kmirror=False,
                altaz_tracking=False,
                force=force,
            )

        if disable:
            self.write_to_log("Disabling telescope.")
            await self.pwi.commands.setEnabled(False)

        if kmirror and self.km:
            self.write_to_log("Parking k-mirror.", level="info")
            await self.km.park()

        self.is_homed = False

    async def stop(self):
        """Stops the mount."""

        self.write_to_log("Stopping the mount.", "warning")
        await self.actor.commands.stop()

    async def goto_coordinates(
        self,
        ra: float | None = None,
        dec: float | None = None,
        pa: float = 0.0,
        alt: float | None = None,
        az: float | None = None,
        kmirror: bool = True,
        kmirror_kwargs: dict = {},
        altaz_tracking: bool = False,
        force: bool = False,
        retry: bool = True,
    ):
        """Moves the telescope to a given RA/Dec or Alt/Az.

        Parameters
        ----------
        ra
            Right ascension coordinates to move to, in degrees.
        dec
            Declination coordinates to move to, in degrees.
        pa
            Position angle of the IFU. Defaults to PA=0.
        alt
            Altitude coordinates to move to, in degrees.
        az
            Azimuth coordinates to move to, in degrees.
        kmirror
            Whether to move the k-mirror into position. Only when
            the coordinates provided are RA/Dec.
        kmirror_kwargs
            Dictionary of keyword arguments to pass to :obj:`.KMirror.slew`.
        altaz_tracking
            If :obj:`True`, starts tracking after moving to alt/az coordinates.
            By default the PWI won't track with those coordinates.
        force
            Move the telescopes even if mode is local.
        retry
            Retry once if the coordinates are not reached.

        """

        if not force and (await self.gort.enclosure.is_local()):
            self.write_to_log("Checking if enclosure is in local mode.")
            raise GortTelescopeError(
                "Cannot move telescope in local mode.",
                error_code=ErrorCode.CANNOT_MOVE_LOCAL_MODE,
            )

        kmirror_task: asyncio.Task | None = None
        if kmirror and self.km and ra is not None and dec is not None:
            kmirror_task = asyncio.create_task(
                self.km.slew(
                    ra,
                    dec,
                    offset_angle=pa,
                    **kmirror_kwargs,
                )
            )

        # Commanded and reported coordinates. To be used to check if we reached
        # the correct position.
        commanded: tuple[float, float]
        reported: tuple[float, float]

        await self.initialise()

        if not self.is_homed:
            self.write_to_log("Telescope is not homed. Homing now.", "warning")
            await self.home()

        if ra is not None and dec is not None:
            is_radec = ra is not None and dec is not None and not alt and not az
            assert is_radec, "Invalid input parameters"

            self.write_to_log(f"Moving to ra={ra:.6f} dec={dec:.6f}.", level="info")

            ra = float(numpy.clip(ra, 0, 360))
            dec = float(numpy.clip(dec, -90, 90))

            await self.pwi.commands.gotoRaDecJ2000(
                ra / 15.0,
                dec,
                timeout=self.timeouts["slew"],
            )

            # Check the position the PWI reports.
            status = await self.status()
            ra_status = status["ra_j2000_hours"] * 15
            dec_status = status["dec_j2000_degs"]

            commanded = (ra, dec)
            reported = (ra_status, dec_status)

        elif alt is not None and az is not None:
            is_altaz = alt is not None and az is not None and not ra and not dec
            assert is_altaz, "Invalid input parameters"

            self.write_to_log(f"Moving to alt={alt:.6f} az={az:.6f}.", level="info")
            await self.pwi.commands.gotoAltAzJ2000(
                alt,
                az,
                timeout=self.timeouts["slew"],
            )

            # Check the position the PWI reports.
            status = await self.status()
            az_status = status["azimuth_degs"]
            alt_status = status["altitude_degs"]

            commanded = (az, alt)
            reported = (az_status, alt_status)

        else:
            raise GortTelescopeError("Invalid coordinates.")

        # Check if we reached the position. If not retry once or fail.
        separation = angular_separation(*commanded, *reported)
        if separation > 0.5:
            if retry:
                self.write_to_log(
                    "Telescope failed to reach the desired position. Retrying",
                    "warning",
                )
                # Need to make sure the k-mirror is not moving before re-trying.
                if kmirror_task is not None and not kmirror_task.done():
                    await kmirror_task

                await asyncio.sleep(3)

                return await self.goto_coordinates(
                    ra=ra,
                    dec=dec,
                    alt=alt,
                    az=az,
                    pa=pa,
                    kmirror=kmirror,
                    kmirror_kwargs=kmirror_kwargs,
                    altaz_tracking=altaz_tracking,
                    force=force,
                    retry=False,
                )
            else:
                await self.actor.commands.setEnabled(False)
                raise GortTelescopeError(
                    "Telescope failed to reach desired position. "
                    "The axes have been disabled for safety. "
                    "Try re-homing the telescope.",
                    error_code=ErrorCode.FAILED_REACHING_COMMANDED_POSITION,
                )

        if alt is not None and az is not None and altaz_tracking:
            await self.pwi.commands.setTracking(enable=True)

        if kmirror_task is not None:
            await kmirror_task

    async def goto_named_position(
        self,
        name: str,
        altaz_tracking: bool = False,
        force: bool = False,
    ):
        """Sends the telescope to a named position.

        Parameters
        ----------
        name
            The name of the position, e.g., ``'zenith'``.
        altaz_tracking
            Whether to start tracking after reaching the position, if the
            coordinates are alt/az.
        force
            Move the telescope even in local enclosure mode.

        """

        if not force and (await self.gort.enclosure.is_local()):
            raise GortTelescopeError(
                "Cannot move telescope in local mode.",
                error_code=ErrorCode.CANNOT_MOVE_LOCAL_MODE,
            )

        if name not in self.config["named_positions"]:
            raise GortTelescopeError(
                f"Invalid named position {name!r}.",
                error_code=ErrorCode.INVALID_TELESCOPE_POSITION,
            )

        position_data = self.config["named_positions"][name]

        if self.name in position_data:
            coords = position_data[self.name]
        elif "all" in position_data:
            coords = position_data["all"]
        else:
            raise GortTelescopeError(
                "Cannot find position data.",
                error_code=ErrorCode.INVALID_TELESCOPE_POSITION,
            )

        if "alt" in coords and "az" in coords:
            coro = self.goto_coordinates(
                alt=coords["alt"],
                az=coords["az"],
                altaz_tracking=altaz_tracking,
                force=force,
            )
        elif "ra" in coords and "dec" in coords:
            coro = self.goto_coordinates(
                ra=coords["ra"],
                dec=coords["dec"],
                force=force,
            )
        else:
            raise GortTelescopeError(
                "No ra/dec or alt/az coordinates found.",
                error_code=ErrorCode.INVALID_TELESCOPE_POSITION,
            )

        await coro

    async def offset(
        self,
        ra: float | None = None,
        dec: float | None = None,
        axis0: float | None = None,
        axis1: float | None = None,
    ):
        """Apply an offset to the telescope axes.

        Parameters
        ----------
        ra
            Offset in RA, in arcsec.
        dec
            Offset in Dec, in arcsec.
        axis0
            Offset in axis 0, in arcsec.
        axis1
            Offset in axis 1, in arcsec.

        """

        if any([ra, dec]) and any([axis0, axis1]):
            raise GortTelescopeError(
                "Cannot offset in ra/dec and axis0/axis1 at the same time."
            )

        kwargs = {}
        if any([ra, dec]):
            if ra is not None:
                kwargs["ra_add_arcsec"] = ra
            if dec is not None:
                kwargs["dec_add_arcsec"] = dec
            self.write_to_log(f"Offsetting RA={ra:.3f}, Dec={dec:.3f}.")

        elif any([axis0, axis1]):
            if axis0 is not None:
                kwargs["axis0_add_arcsec"] = axis0
            if axis1 is not None:
                kwargs["axis1_add_arcsec"] = axis1
            self.write_to_log(f"Offsetting axis0={axis0:.3f}, axis1={axis1:.3f}.")

        else:
            # This should not happen.
            raise GortTelescopeError("No offsets provided.")

        await self.actor.commands.offset(**kwargs)

        self.write_to_log("Offset complete.")

        return True


class TelescopeSet(GortDeviceSet[Telescope]):
    """A representation of a set of telescopes."""

    __DEVICE_CLASS__ = Telescope
    __DEPLOYMENTS__ = [
        "lvmpwi-sci",
        "lvmpwi-spec",
        "lvmpwi-skye",
        "lvmpwi-skyw",
        "lvmtan",
    ]

    def __init__(self, gort: Gort, data: dict[str, dict]):
        super().__init__(gort, data)

        self.guiders = self.gort.guiders

    async def initialise(self):
        """Initialise all telescopes."""

        await asyncio.gather(*[tel.initialise() for tel in self.values()])

    async def restart(self):
        """Restarts the ``lvmpwi`` and ``lvmtan`` deployments and re-homes."""

        result = await super().restart()

        if result is False:
            self.write_to_log(
                "Some deployments failed to restart. Not homing devices.",
                "error",
            )
            return

        self.write_to_log("Waiting 10 seconds for devices to reconnect.", "info")
        await asyncio.sleep(10)

        self.write_to_log("Homing telescope a restart.")
        await self.home(home_kms=True, home_focusers=True, home_fibsel=True)

    async def restart_lvmtan(self):
        """Restarts and rehomes Twice-As-Nice controller.

        After the actor has been restarted the K-mirrors, focuser, and fibre selector
        are rehomed. The focuser positions are preserved.

        """

        self.write_to_log("Restarting deployment lvmtan and waiting 25 s.", "info")
        await kubernetes_restart_deployment("lvmtan")
        await asyncio.sleep(25)

        await self.home(
            home_telescopes=False,
            home_kms=True,
            home_focusers=True,
            home_fibsel=True,
        )

    async def home(
        self,
        home_telescopes: bool = True,
        home_kms: bool = True,
        home_focusers: bool = False,
        home_fibsel: bool = False,
    ):
        """Initialises and homes the telescope.

        Parameters
        ---------
        home_telescopes
            Homes the telescopes. Defaults to :obj:`True`.
        home_kms
            Homes the K-mirrors, if present. Defaults to :obj:`True`.
        home_focusers
            Homes the focusers. Defaults to :obj:`False`.
        home_fibsel
            Homes the fibre selector, if present. Defaults to :obj:`False`.

        """

        self.write_to_log("Rehoming all telescopes.", "info")

        await self.call_device_method(
            Telescope.home,
            home_telescope=home_telescopes,
            home_km=home_kms,
            home_focuser=home_focusers,
            home_fibsel=home_fibsel,
        )

    async def park(
        self,
        disable=True,
        use_pw_park=False,
        alt_az: tuple[float, float] | None = None,
        kmirror=True,
        force=False,
    ):
        """Parks the telescopes.

        Parameters
        ----------
        disable
            Disables the axes after reaching the park position.
        use_pw_park
            Uses the internal park position stored in the PlaneWave mounts.
        alt_az
            A tuple with the alt and az position at which to park the telescopes.
            If not provided, defaults to the ``park`` named position.
        kmirror
            Whether to park the mirrors at 90 degrees.
        force
            Moves the telescopes even if the mode is local.

        """

        await self.call_device_method(
            Telescope.park,
            disable=disable,
            use_pw_park=use_pw_park,
            alt_az=alt_az,
            kmirror=kmirror,
            force=force,
        )

    async def goto_coordinates_all(
        self,
        ra: float | None = None,
        dec: float | None = None,
        alt: float | None = None,
        az: float | None = None,
        kmirror: bool = True,
        altaz_tracking: bool = False,
        force: bool = False,
    ):
        """Moves all the telescopes to a given RA/Dec or Alt/Az.

        Parameters
        ----------
        ra
            Right ascension coordinates to move to.
        dec
            Declination coordinates to move to.
        alt
            Altitude coordinates to move to.
        az
            Azimuth coordinates to move to.
        kmirror
            Whether to move the k-mirror into position. Only when
            the coordinates provided are RA/Dec.
        altaz_tracking
            If :obj:`True`, starts tracking after moving to alt/az coordinates.
            By defaul the PWI won't track with those coordinates.
        force
            Move the telescopes even if mode is local.

        """

        await self.call_device_method(
            Telescope.goto_coordinates,
            ra=ra,
            dec=dec,
            alt=alt,
            az=az,
            kmirror=kmirror,
            altaz_tracking=altaz_tracking,
            force=force,
        )

    async def goto_named_position(
        self,
        name: str,
        altaz_tracking: bool = False,
        force: bool = False,
    ):
        """Sends the telescopes to a named position.

        Parameters
        ----------
        name
            The name of the position, e.g., ``'zenith'``.
        altaz_tracking
            Whether to start tracking after reaching the position, if the
            coordinates are alt/az.
        force
            Move the telescopes even in local enclosure mode.

        """

        await self._check_local(force)

        if name not in self.gort.config["telescopes"]["named_positions"]:
            raise GortTelescopeError(
                f"Invalid named position {name!r}.",
                error_code=ErrorCode.INVALID_TELESCOPE_POSITION,
            )

        await self.call_device_method(
            Telescope.goto_named_position,
            name=name,
            altaz_tracking=altaz_tracking,
            force=force,
        )

    async def stop(self):
        """Stops all the mounts."""

        await self.call_device_method(Telescope.stop)

    async def goto(
        self,
        sci: tuple[float, float] | tuple[float, float, float] | None = None,
        spec: tuple[float, float] | None = None,
        skye: tuple[float, float] | None = None,
        skyw: tuple[float, float] | None = None,
        sci_km_stop_degs_before: float = 0.0,
        force: bool = False,
    ):
        """Sends each telescope to a different position.

        Parameters
        ----------
        sci
            The RA and Dec where to slew the science telescope. A third value
            can be provided for the PA. Note that in this case a -1 factor
            is applied before sending it to the K-mirror.
        spec
            The RA and Dec where to slew the spectrophotometric telescope.
        skye
            The RA and Dec where to slew the skyE telescope.
        skyw
            The RA and Dec where to slew the skyW telescope.
        sci_km_stop_degs_before
            The number of degrees before the desired position where to send
            the science k-mirror. Useful if we want to be sure that positive
            offsets will be applied without backlash.

        """

        await self._check_local(force)

        jobs = []

        if sci is not None:
            kmirror_kwargs = {"stop_degs_before": sci_km_stop_degs_before}
            if len(sci) == 2:
                jobs.append(
                    self["sci"].goto_coordinates(
                        ra=sci[0],
                        dec=sci[1],
                        kmirror_kwargs=kmirror_kwargs,
                    )
                )
            else:
                jobs.append(
                    self["sci"].goto_coordinates(
                        ra=sci[0],
                        dec=sci[1],
                        pa=-sci[2],  # Note the -1 here
                        kmirror_kwargs=kmirror_kwargs,
                    )
                )

        if spec is not None:
            jobs.append(self["spec"].goto_coordinates(ra=spec[0], dec=spec[1]))

        if skye is not None:
            jobs.append(self["skye"].goto_coordinates(ra=skye[0], dec=skye[1]))

        if skyw is not None:
            jobs.append(self["skyw"].goto_coordinates(ra=skyw[0], dec=skyw[1]))

        if len(jobs) == 0:
            return

        await asyncio.gather(*jobs)

    async def _check_local(self, force: bool = False):
        """Checks if the telescope is in local mode and raises an error."""

        is_local = await self.gort.enclosure.is_local()
        if is_local and not force:
            raise GortTelescopeError(
                "Cannot move telescopes in local mode.",
                error_code=ErrorCode.CANNOT_MOVE_LOCAL_MODE,
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-26
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)


from __future__ import annotations

from .alerts import AlertsOverwatcher
from .calibration import CalibrationsOverwatcher
from .core import OverwatcherModule
from .ephemeris import EphemerisOverwatcher
from .events import EventsOverwatcher
from .helpers import DomeHelper, DomeStatus, NotifierMixIn
from .observer import ObserverOverwatcher
from .overwatcher import Overwatcher
from .safety import SafetyOverwatcher
from .transparency import TransparencyOverwatcher
from .weather import WeatherOverwatcher



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-10-27
# @Filename: alerts.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import enum
from time import time

from typing import TYPE_CHECKING

from lvmopstools.retrier import Retrier
from pydantic import BaseModel

from gort.overwatcher.core import OverwatcherModuleTask
from gort.overwatcher.overwatcher import OverwatcherModule
from gort.tools import get_lvmapi_route


if TYPE_CHECKING:
    pass


__all__ = ["AlertsOverwatcher", "ActiveAlert"]


class AlertsSummary(BaseModel):
    """Summary of alerts."""

    humidity_alert: bool | None = None
    dew_point_alert: bool | None = None
    wind_alert: bool | None = None
    rain: bool | None = None
    door_alert: bool | None = None
    camera_temperature_alert: bool | None = None
    camera_alerts: dict[str, bool] | None = None
    o2_alert: bool | None = None
    o2_room_alerts: dict[str, bool] | None = None
    heater_alert: bool | None = None
    heater_camera_alerts: dict[str, bool] | None = None


class ActiveAlert(enum.Flag):
    """Flags for active alerts."""

    HUMIDITY = enum.auto()
    DEW_POINT = enum.auto()
    WIND = enum.auto()
    RAIN = enum.auto()
    DOOR = enum.auto()
    CAMERA_TEMPERATURE = enum.auto()
    O2 = enum.auto()
    LOCKED = enum.auto()
    UNKNOWN = enum.auto()


class AlertsMonitorTask(OverwatcherModuleTask["AlertsOverwatcher"]):
    """Monitors the alerts state."""

    name = "alerts_monitor"
    keep_alive = True
    restart_on_error = True

    def __init__(self):
        super().__init__()

        self.last_updated: float = 0.0
        self.unavailable: bool = False

    async def task(self):
        """Updates the alerts data."""

        n_failures: int = 0

        while True:
            try:
                await self.update_alerts()
            except Exception as err:
                if self.unavailable is False:
                    self.log.error(f"Failed to get alerts data: {err!r}")
                n_failures += 1
            else:
                self.last_updated = time()
                self.unavailable = False
                n_failures = 0
            finally:
                if self.unavailable is False and n_failures >= 5:
                    self.unavailable = True
                    self.log.critical(
                        "Failed to get alerts data 5 times. "
                        "Triggering an emergency shutdown.",
                    )
                    asyncio.create_task(
                        self.overwatcher.shutdown(
                            reason="alerts data unavailable",
                            park=False,
                        )
                    )

            await asyncio.sleep(15)

    async def update_alerts(self):
        """Processes the weather update and determines whether it is safe to observe."""

        data = await self.module.get_alerts_summary()

        if data is None:
            raise ValueError("no alerts data available.")

        # For some very critical alerts, we require them to be not null (null here
        # means no data was available or the API failed getting the alert data).
        if data.rain is None or data.humidity_alert is None or data.wind_alert is None:
            raise ValueError("incomplete alerts data.")

        self.module.state = data


class AlertsOverwatcher(OverwatcherModule):
    """Monitors alerts."""

    name = "alerts"

    tasks = [AlertsMonitorTask()]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.state: AlertsSummary | None = None
        self.locked_until: float = 0

    def is_safe(self) -> tuple[bool, ActiveAlert]:
        """Determines whether it is safe to open."""

        if self.state is None:
            self.log.warning("Alerts data not available. is_safe() returns False.")
            return False, ActiveAlert.UNKNOWN

        is_safe: bool = True
        active_alerts = ActiveAlert(0)

        if self.state.rain:
            self.log.warning("Rain alert detected.")
            active_alerts |= ActiveAlert.RAIN
            is_safe = False
        if self.state.humidity_alert:
            self.log.warning("Humidity alert detected.")
            active_alerts |= ActiveAlert.HUMIDITY
            is_safe = False
        if self.state.dew_point_alert:
            self.log.warning("Dew point alert detected.")
            active_alerts |= ActiveAlert.DEW_POINT
            is_safe = False
        if self.state.wind_alert:
            self.log.warning("Wind alert detected.")
            active_alerts |= ActiveAlert.WIND
            is_safe = False

        # These alerts are not critical but we log them.
        # TODO: maybe we do want to do something about these alerts.
        if self.state.door_alert:
            active_alerts |= ActiveAlert.DOOR
        if self.state.camera_temperature_alert:
            self.log.warning("Camera temperature alert detected.")
            active_alerts |= ActiveAlert.CAMERA_TEMPERATURE
        if self.state.o2_alert:
            self.log.warning("O2 alert detected.")
            active_alerts |= ActiveAlert.O2

        # If we have issued a previous unsafe alert, the main task will close the dome
        # and put a lock for 30 minutes to prevent the dome from opening/closing too
        # frequently if the weather is unstable.
        if self.locked_until > 0 and time() < self.locked_until:
            return False, active_alerts | ActiveAlert.LOCKED

        if is_safe:
            self.locked_until = 0

        return is_safe, active_alerts

    @Retrier(max_attempts=3, delay=5)
    async def get_alerts_summary(self) -> AlertsSummary:
        """Returns the alerts report."""

        data = await get_lvmapi_route("/alerts/summary")

        return AlertsSummary(**data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-26
# @Filename: core.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from unittest.mock import Mock
from weakref import WeakSet

from typing import TYPE_CHECKING, ClassVar, Generic, TypeVar

from gort.exceptions import GortError
from gort.tools import LogNamespace, cancel_task


if TYPE_CHECKING:
    from gort.overwatcher.overwatcher import Overwatcher


OverwatcherModule_T = TypeVar("OverwatcherModule_T", bound="OverwatcherModule")


class OverwatcherBaseTask:
    """A task that runs in an overwatcher module."""

    name: ClassVar[str]
    keep_alive: ClassVar[bool] = True
    restart_on_error: ClassVar[bool] = True

    def __init__(self, log: LogNamespace | None = None):
        self._task_runner: asyncio.Task | None = None
        self._heartbeat_task: asyncio.Task | None = None

        self._log: Mock | LogNamespace = log or Mock()

    async def run(self):
        """Runs the task."""

        self._task_runner = asyncio.create_task(self.task())
        self._log.debug(f"Task {self.name!r} started.")

        self._heartbeat_task = asyncio.create_task(self.monitor_task_heartbeat())

    async def cancel(self):
        """Cancels the task."""

        await cancel_task(self._task_runner)
        await cancel_task(self._heartbeat_task)

        self._task_runner = None
        self._heartbeat_task = None

        self._log.debug(f"Task {self.name!r} was cancelled.")

    async def task(self):
        """The task to run."""

        raise NotImplementedError("The task coroutine has not been implemented.")

    async def monitor_task_heartbeat(self):
        """Monitors the task and restarts it if it fails."""

        while True:
            await asyncio.sleep(1)

            if not self._task_runner:
                continue

            if self._task_runner.cancelled():
                if self.keep_alive:
                    self._log.warning(f"Task {self.name!r} was cancelled. Restarting.")
                    self._task_runner = asyncio.create_task(self.task())
                return

            if self._task_runner.done():
                if exception := self._task_runner.exception():
                    self._log.error(f"Task {self.name!r} failed: {exception!r}")
                    if self.restart_on_error or self.keep_alive:
                        self._log.warning(f"Task {self.name!r} failed. Restarting.")
                        self._task_runner = asyncio.create_task(self.task())
                        continue

                if self.keep_alive:
                    self._log.warning(f"Task {self.name!r} finished. Restarting.")
                    self._task_runner = asyncio.create_task(self.task())
                    continue


class OverwatcherModuleTask(OverwatcherBaseTask, Generic[OverwatcherModule_T]):
    """A task that runs in an overwatcher module."""

    def __init__(self):
        super().__init__()

        self._module: OverwatcherModule_T | None = None

    @property
    def module(self):
        """Returns the module instance."""

        if not self._module:
            raise GortError("Task has not been associated with a module.")

        return self._module

    @property
    def overwatcher(self):
        """Returns the overwatcher instance."""

        return self.module.overwatcher

    @property
    def gort(self):
        """Returns the GORT instance."""

        return self.module.gort

    @property
    def config(self):
        """Returns the configuration dictionary."""

        return self.module.gort.config

    @property
    def log(self):
        """Returns the logger instance."""

        return self.module.log

    @property
    def notify(self):
        """Returns the Overwatcher notifier."""

        return self.module.overwatcher.notify

    async def run(self, module: OverwatcherModule_T):
        """Runs the task."""

        self._module = module

        await super().run()


class OverwatcherModule:
    """A generic overwatcher module."""

    instances = WeakSet()
    name: str = "generic"
    delay: float = 0

    tasks: ClassVar[list[OverwatcherModuleTask]]

    def __new__(cls, *args, **kwargs):
        instance = object.__new__(cls)
        cls.instances.add(instance)

        return instance

    def __init__(self, overwatcher: Overwatcher):
        if not hasattr(self, "tasks"):
            raise RuntimeError(
                f"{self.__class__.__name__} must define a `tasks` attribute."
            )

        self.overwatcher = overwatcher
        self.gort = overwatcher.gort
        self.notify = overwatcher.notify

        self.is_running: bool = False

        self.log = LogNamespace(self.gort.log, header=f"({self.__class__.__name__}) ")
        for task in self.tasks:
            task._log = self.log

    async def run(self):
        """Runs the overwatcher module."""

        if self.is_running:
            raise GortError(f"{self.name!r} overwatcher is already running.")

        await asyncio.sleep(self.delay)

        self.is_running = True

        for ov_task in self.tasks:
            await ov_task.run(self)

        return self

    async def cancel(self):
        """Stops the overwatcher module."""

        for ov_task in self.tasks:
            await ov_task.cancel()

        self.is_running = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-18
# @Filename: enclosure.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING

from lvmopstools import Retrier

from gort.devices.core import GortDevice, GortDeviceSet
from gort.enums import Event
from gort.exceptions import ErrorCode, GortEnclosureError, GortTelescopeError


if TYPE_CHECKING:
    from gort import Gort


if TYPE_CHECKING:
    from gort import ActorReply


__all__ = ["Enclosure", "Lights", "Light"]


class Light:
    """An enclosure light."""

    def __init__(self, enclosure: Enclosure, name: str):
        self.enclosure = enclosure
        self.name = name

    @Retrier(max_attempts=3, delay=1)
    async def on(self):
        """Turns on the light."""

        status = await self.status()
        if status is False:
            self.enclosure.write_to_log(f"Turning on light {self.name!r}.", "info")
            await self.enclosure.actor.commands.lights("on", self.name)

    @Retrier(max_attempts=3, delay=1)
    async def off(self):
        """Turns on the light."""

        status = await self.status()
        if status is True:
            self.enclosure.write_to_log(f"Turning off light {self.name!r}.", "info")
            await self.enclosure.actor.commands.lights("off", self.name)

    @Retrier(max_attempts=3, delay=1)
    async def toggle(self):
        """Turns on the light."""

        self.enclosure.write_to_log(f"Toggling {self.name!r}.", "info")
        await self.enclosure.actor.commands.lights("toggle", self.name)

    @Retrier(max_attempts=3, delay=1)
    async def status(self) -> bool:
        """Returns a boolean with the light status."""

        status = await self.enclosure.actor.commands.lights("status")

        labels = status.get("lights_labels")

        if labels is None:
            raise GortEnclosureError("Did not receive lights status.")

        return self.name.upper() in labels


class Lights:
    """Controls the enclosure lights."""

    LIGHTS = [
        "telescope_bright",
        "telescope_red",
        "spectrograph_room",
        "utilities_room",
    ]

    def __init__(self, enclosure: Enclosure):
        for light in self.LIGHTS:
            self.__setattr__(light, Light(enclosure, light))

    def __repr__(self):
        return f"<Lights ({', '.join(self.LIGHTS)})>"

    @Retrier(max_attempts=3, delay=1)
    async def dome_all_off(self):
        """Turns off all the lights in the dome."""

        await asyncio.gather(self.telescope_bright.off(), self.telescope_red.off())

    if TYPE_CHECKING:

        def __getattr__(self, light: str) -> Light: ...


class Enclosure(GortDevice):
    """Class representing the LVM enclosure."""

    __DEPLOYMENTS__ = ["lvmecp"]

    def __init__(self, gort: Gort, name: str, actor: str, **kwargs):
        super().__init__(gort, name, actor)

        self.lights = Lights(self)

    async def restart(self):
        """Restarts the ``lvmecp`` deployment."""

        await GortDeviceSet.restart(self)  # type: ignore

    @Retrier(max_attempts=3, delay=1)
    async def status(self, get_registers: bool = False):
        """Retrieves the status of the power outlet."""

        reply: ActorReply = await self.actor.commands.status(
            no_registers=(not get_registers),
            timeout=5,
        )

        return reply.flatten()

    async def _prepare_telescopes(self, force: bool = False):
        """Moves telescopes to park position before opening/closing the enclosure."""

        telescopes = list(self.gort.telescopes)
        is_parked = await asyncio.gather(
            *[self.gort.telescopes[tel].is_parked() for tel in telescopes]
        )

        if all(is_parked):
            return True

        self.write_to_log(
            "Moving telescopes to park before operating the dome.",
            "warning",
        )

        # Check local only once here.
        if await self.gort.enclosure.is_local():
            raise GortTelescopeError(
                "Cannot move telescope in local mode.",
                error_code=ErrorCode.CANNOT_MOVE_LOCAL_MODE,
            )

        park_coros = [
            self.gort.telescopes[tel].goto_named_position("park", force=True)
            for itel, tel in enumerate(telescopes)
            if not is_parked[itel]
        ]

        await asyncio.gather(*park_coros)

    async def open(self, park_telescopes: bool = True):
        """Open the enclosure dome.

        Parameters
        ----------
        park_telescopes
            Move the telescopes to the park position before opening the
            enclosure to prevent dust or other debris falling on them.

        """

        is_local = await self.is_local()
        if is_local:
            raise GortEnclosureError(
                "Cannot open the enclosure while in local mode.",
                error_code=ErrorCode.LOCAL_MODE_FAILED,
            )

        if park_telescopes:
            await self._prepare_telescopes()

        self.write_to_log("Opening the enclosure ...", level="info")
        await self.gort.notify_event(Event.DOME_OPENING)

        await self.actor.commands.dome.commands.open()

        self.write_to_log("Enclosure is now open.", level="info")
        await self.gort.notify_event(Event.DOME_OPEN)

    async def close(
        self,
        park_telescopes: bool = True,
        force: bool = False,
        retry_without_parking: bool = False,
    ):
        """Close the enclosure dome.

        Parameters
        ----------
        park_telescopes
            Move the telescopes to the park position before closing the
            enclosure to prevent dust or other debris falling on them.
        force
            Tries to closes the dome even if the system believes it is
            already closed.
        retry_without_parking
            If the dome fails to close with ``park_telescopes=True``, it will
            try again without parking the telescopes.

        """

        is_local = await self.is_local()
        if is_local:
            raise GortEnclosureError(
                "Cannot close the enclosure while in local mode.",
                error_code=ErrorCode.LOCAL_MODE_FAILED,
            )

        self.write_to_log("Closing the dome ...", level="info")
        await self.gort.notify_event(Event.DOME_CLOSING)

        try:
            if park_telescopes:
                try:
                    await self._prepare_telescopes(force=force)
                except Exception as err:
                    self.write_to_log(
                        f"Failed determining the status of the telescopes: {err}",
                        "warning",
                    )
                    if force is False:
                        raise GortEnclosureError(
                            "Not closing without knowing where the telescopes are. "
                            "If you really need to close call again with force=True."
                        )
                    else:
                        self.write_to_log("Closing because force=True", "warning")

            await self.actor.commands.dome.commands.close(force=force)

        except Exception as err:
            if retry_without_parking is False or park_telescopes is False:
                raise

            self.write_to_log(f"Failed to close the dome: {err}", "warning")
            self.write_to_log("Retrying without parking the telescopes.", "warning")
            await self.close(park_telescopes=False, force=force)

        self.write_to_log("Enclosure is now closed.", level="info")
        await self.gort.notify_event(Event.DOME_CLOSED)

    async def is_open(self):
        """Returns :obj:`True` if the enclosure is open."""

        status = await self.status()
        labels = status["dome_status_labels"]

        return "OPEN" in labels and "MOVING" not in labels

    async def is_closed(self):
        """Returns :obj:`True` if the enclosure is closed."""

        status = await self.status()
        labels = status["dome_status_labels"]

        return "CLOSED" in labels and "MOVING" not in labels

    @Retrier(max_attempts=2, delay=1)
    async def stop(self):
        """Stop the enclosure dome."""

        self.write_to_log("Stoping the dome.", level="info")
        await self.actor.commands.dome.commands.stop()

    async def is_local(self):
        """Returns :obj:`True` if the enclosure is in local mode."""

        # This should generally not be on, but it's useful as a way of disabling
        # the local mode when the lock or door are not working.
        if self.gort.config["enclosure"].get("bypass_local_mode", False) is True:
            return False

        status = await self.status()
        safety_status_labels = status.get("safety_status_labels", None)
        if safety_status_labels is None:
            raise GortEnclosureError(
                "Cannot determine if enclosure is in local mode.",
                error_code=ErrorCode.LOCAL_MODE_FAILED,
            )

        return "LOCAL" in safety_status_labels

    async def get_door_status(self):
        """Returns the status of the door and lock."""

        status = await self.status()
        safety_status_labels = status.get("safety_status_labels", None)
        if safety_status_labels is None:
            raise GortEnclosureError(
                "Cannot determine door status.",
                error_code=ErrorCode.DOOR_STATUS_FAILED,
            )

        reply = {
            "door_closed": "DOOR_CLOSED" in safety_status_labels,
            "door_locked": "DOOR_LOCKED" in safety_status_labels,
            "local": await self.is_local(),
        }

        return reply



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-16
# @Filename: gort.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging
import os
import pathlib
import subprocess
import sys
import uuid
from copy import deepcopy
from functools import partial
from types import TracebackType

from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Type,
    TypeVar,
)

from lvmopstools.retrier import Retrier
from rich import pretty, traceback
from rich.logging import RichHandler
from typing_extensions import Self

from clu.client import AMQPClient
from sdsstools.logger import SDSSLogger, get_logger
from sdsstools.time import get_sjd

from gort import config
from gort.devices.core import GortDevice, GortDeviceSet
from gort.enums import Event
from gort.exceptions import ErrorCode, GortError
from gort.pubsub import notify_event
from gort.recipes import recipes as recipe_to_class
from gort.remote import RemoteActor
from gort.tile import Tile
from gort.tools import (
    get_temporary_file_path,
    kubernetes_restart_deployment,
    overwatcher_is_running,
    run_in_executor,
    set_tile_status,
)


if TYPE_CHECKING:
    from rich.console import Console


try:
    IPYTHON = get_ipython()  # type: ignore
    IPYTHON_DEFAULT_HOOKS = [
        IPYTHON._showtraceback,
        IPYTHON.showtraceback,
        IPYTHON.showsyntaxerror,
    ]
except NameError:
    IPYTHON = None
    IPYTHON_DEFAULT_HOOKS = []


__all__ = ["GortClient", "Gort"]


DevType = TypeVar("DevType", bound="GortDeviceSet | GortDevice")


class GortClient(AMQPClient):
    """The main ``gort`` client, used to communicate with the actor system.

    A subclass of :obj:`~clu.client.AMQPClient` with defaults for host and logging,
    it loads the :obj:`.GortDeviceSet` and :obj:`.RemoteActor` instances
    for the LVM system.

    Parameters
    ----------
    host
        The host on which the RabbitMQ exchange is running.
    port
        The port on which the RabbitMQ exchange listens to connections.
    user
        The user to connect to the exchange.
    password
        The password to connect to the exchange.
    use_rich_output
        If :obj:`True`, uses ``rich`` to provide colourised tracebacks and
        prettier outputs.
    log_file_path
        The path where to save GORT's log. File logs are always saved with ``DEBUG``
        logging level. If :obj:`None`, a temporary file will be used whose path
        can be retrieved by calling :obj:`.get_log_path`. If :obj:`False`, no file
        logging will happen.

    """

    def __init__(
        self,
        host: str = "lvm-hub.lco.cl",
        port: int = 5672,
        user: str = "guest",
        password: str = "guest",
        use_rich_output: bool = True,
        log_file_path: str | pathlib.Path | Literal[False] | None = None,
    ):
        self.client_uuid = str(uuid.uuid4()).split("-")[0]

        self._console: Console

        log = self._prepare_logger(
            log_file_path,
            use_rich_output=use_rich_output,
        )

        self._setup_exception_hooks(log, use_rich_output=use_rich_output)

        self._connect_lock = asyncio.Lock()

        super().__init__(
            f"Gort-client-{self.client_uuid}",
            host=host,
            port=port,
            user=user,
            password=password,
            log=log,
        )

    def _prepare_logger(
        self,
        log_file_path: str | pathlib.Path | Literal[False] | None = None,
        use_rich_output: bool = True,
    ):
        """Creates a logger and start file logging."""

        log = get_logger(
            f"lvmgort-{self.client_uuid}",
            use_rich_handler=True,
            rich_handler_kwargs={"rich_tracebacks": use_rich_output},
        )

        if log_file_path is None:
            path = config["logging"]["path"]
            sjd = get_sjd()
            path = path.format(SJD=sjd)

            try:
                log.start_file_logger(str(path), rotating=False, mode="a")
                log.info("Starting GORT.")
                log.info(f"Logging to {str(path)}")

            except Exception as err:
                tmp_path = get_temporary_file_path(
                    prefix="gort-",
                    suffix=f"-{self.client_uuid}.log",
                )
                log.warning(
                    f"Failed starting file logging to {str(path)}: {err}. "
                    f"Using temporary path {str(tmp_path)}."
                )

                log.start_file_logger(str(tmp_path), rotating=False)

        elif log_file_path is not False:
            log.start_file_logger(str(log_file_path), rotating=False)

        assert isinstance(log.sh, RichHandler)
        self._console = log.sh.console

        return log

    async def notify_event(self, event: Event, payload: dict[str, Any] = {}):
        """Emits an event notification."""

        try:
            await notify_event(event, payload=payload)
        except TypeError as err:
            if "is not JSON serializable" in str(err):
                self.log.error(
                    f"Failed to notify event {event.name}: payload is not"
                    "serialisable. The event will be emitted without payload."
                )
                await notify_event(event)

    def exception_handler(
        self,
        log: SDSSLogger,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ):
        """A custom exception handler that logs exceptions to file."""

        log.handle_exceptions(exc_type, exc_value, exc_traceback)

        if exc_value and isinstance(exc_value, GortError):
            event_payload = exc_value.payload.copy()
            event_payload["error"] = exc_value.args[0] or ""
            event_payload["error_code"] = exc_value.error_code.value

            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:  # No running event loop, mostly for IPython.
                log.warning("Cannot emit error event: no running event loop.")
            else:
                loop.create_task(self.notify_event(Event.ERROR, payload=event_payload))

    def asyncio_exception_handler(self, loop, context):
        """Handle an uncaught asyncio exception and reports it."""

        if exception := context.get("exception", None):
            try:
                raise exception
            except Exception:
                exc_type, exc_value, exc_tb = sys.exc_info()
                self.exception_handler(self.log, exc_type, exc_value, exc_tb)
        else:
            loop.default_exception_handler(context)

    def _setup_exception_hooks(
        self,
        log: SDSSLogger | None = None,
        use_rich_output: bool = True,
    ):
        """Setup various hooks for exception handling."""

        log = log or self.log

        def custom__showtraceback_closure(default__showtraceback):
            def _showtraceback(*args, **kwargs):
                assert IPYTHON

                exc_tuple = IPYTHON._get_exc_info()
                self.exception_handler(log, *exc_tuple)

            if IPYTHON:
                IPYTHON._showtraceback = _showtraceback

        if use_rich_output:
            if not IPYTHON:
                traceback.install(console=self._console)
                pretty.install(console=self._console)

            # traceback.install() overrides the excepthook, which means that
            # tracebacks are not logged to file anymore. Restore that.
            sys.excepthook = partial(self.exception_handler, log)

        else:
            # Make sure we are not using rich tracebacks anymore, in
            # case we installed them at some point.
            # See https://github.com/Textualize/rich/pull/2972/files

            sys.excepthook = partial(self.exception_handler, log)

            if IPYTHON:
                IPYTHON._showtraceback = IPYTHON_DEFAULT_HOOKS[0]
                IPYTHON.showtraceback = IPYTHON_DEFAULT_HOOKS[1]
                IPYTHON.showsyntaxerror = IPYTHON_DEFAULT_HOOKS[2]

        if IPYTHON:
            # One more override. We want that any exceptions raised in IPython
            # also gets logged to file, but IPython overrides excepthook completely
            # so here we make a custom call to log.handle_exceptions() and then
            # just let it do whatever it was its default (whether that means it
            # was overridden by rich or not).
            custom__showtraceback_closure(IPYTHON._showtraceback)

    def _setup_async_exception_hooks(self):
        """Sets up a custom async exception handler."""

        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:  # No running event loop, mostly for IPython.
            pass
        else:
            loop.set_exception_handler(self.asyncio_exception_handler)

    def get_log_path(self):
        """Returns the path of the log file. :obj:`None` if not logging to file."""

        return self.log.log_filename

    async def init(self) -> Self:
        """Initialises the client.

        Returns
        -------
        object
            The same instance of :obj:`.GortClient` after initialisation.

        """

        if not self.connected:
            async with self._connect_lock:
                await self.start()

        # Override the asyncio exception handler to catch errors in tasks.
        # We do this after AMQPClient.start() because it sets its own exception
        # handler.
        self._setup_async_exception_hooks()

        return self

    @property
    def connected(self):
        """Returns :obj:`True` if the client is connected."""

        return self.connection and self.connection.connection is not None

    def set_verbosity(self, verbosity: str | int | None = None):
        """Sets the level of verbosity to ``debug``, ``info``, or ``warning``.

        Parameters
        ----------
        verbosity
            The level of verbosity. Can be a string level name, an integer, or
            :obj:`None`, in which case the default verbosity will be used.

        """

        verbosity = verbosity or "warning"
        if isinstance(verbosity, int):
            verbosity = logging.getLevelName(verbosity)

        assert isinstance(verbosity, str)

        verbosity = verbosity.lower()
        if verbosity not in ["debug", "info", "warning"]:
            raise ValueError("Invalid verbosity value.")

        level_mapping = logging.getLevelNamesMapping()
        if verbosity_level := level_mapping.get(verbosity.upper()):
            self.log.sh.setLevel(verbosity_level)


class Gort(GortClient):
    """Gort's robotic functionality.

    :obj:`.Gort` is subclass of :obj:`.GortClient` that implements higher-level
    robotic functionality. This is the class a user will normally instantiate and
    interact with.

    Parameters
    ----------
    args, kwargs
        Arguments to pass to :obj:`.GortClient`.
    allow_overwatcher
        If the Overwatcher is running an exception will be raised to prevent
        GORT running commands that may interfere with it. If ``allow_overwatcher=True``
        the exception will be suppressed.
    verbosity
        The level of logging verbosity.

    """

    def __init__(
        self,
        *args,
        override_overwatcher: bool | None = None,
        verbosity: str | None = None,
        config_file: str | pathlib.Path | None = None,
        **kwargs,
    ):
        # Not a circular import issue, but here so that importing Gort is a bit faster.
        from gort.devices.ag import AGSet
        from gort.devices.enclosure import Enclosure
        from gort.devices.guider import GuiderSet
        from gort.devices.nps import NPSSet
        from gort.devices.spec import SpectrographSet
        from gort.devices.telemetry import TelemetrySet
        from gort.devices.telescope import TelescopeSet
        from gort.observer import GortObserver

        if config_file:
            config.load(str(config_file))

        super().__init__(*args, **kwargs)

        self.actors: dict[str, RemoteActor] = {}

        self.config = deepcopy(config)

        self.__device_sets = []

        self.ags = self.add_device(AGSet, config["ags.devices"])
        self.guiders = self.add_device(GuiderSet, config["guiders.devices"])
        self.telescopes = self.add_device(TelescopeSet, config["telescopes.devices"])
        self.nps = self.add_device(NPSSet, config["nps.devices"])
        self.specs = self.add_device(SpectrographSet, config["specs.devices"])
        self.enclosure = self.add_device(Enclosure, name="enclosure", actor="lvmecp")
        self.telemetry = self.add_device(TelemetrySet, config["telemetry.devices"])

        self._override_overwatcher = override_overwatcher

        self.observer = GortObserver(self)
        self.observe_tile = self.observer.observe_tile

        if verbosity:
            self.set_verbosity(verbosity)

    async def init(self) -> Self:
        """Initialises the client and all devices."""

        await super().init()

        override_overwatcher = self._override_overwatcher
        override_envvar = os.environ.get("GORT_OVERRIDE_OVERWATCHER", "0")

        if override_overwatcher is None:
            override_overwatcher = False if override_envvar == "0" else True

        if override_overwatcher is None:
            pass
        else:
            overwatcher_running = await overwatcher_is_running(self)
            if overwatcher_running:
                if override_overwatcher:
                    self.log.warning("Overwatcher is running, be careful!")
                else:
                    raise GortError(
                        "Overwatcher is running. If you really want to use GORT "
                        "initialise it with Gort(override_overwatcher=True).",
                        error_code=ErrorCode.OVERATCHER_RUNNING,
                    )

        await asyncio.gather(*[ractor.init() for ractor in self.actors.values()])

        # Initialise device sets.
        await asyncio.gather(*[dev.init() for dev in self.__device_sets])

        return self

    def add_device(self, class_: Type[DevType], *args, **kwargs) -> DevType:
        """Adds a new device or device set to Gort."""

        ds = class_(self, *args, **kwargs)
        self.__device_sets.append(ds)

        return ds

    def add_actor(self, actor: str, device: GortDevice | None = None):
        """Adds an actor to the programmatic API.

        Parameters
        ----------
        actor
            The name of the actor to add.
        device
            A device associated with this actor.

        """

        if actor not in self.actors:
            self.actors[actor] = RemoteActor(self, actor, device=device)

        return self.actors[actor]

    @Retrier(max_attempts=3, delay=3)
    async def emergency_shutdown(self):
        """Parks and closes the telescopes."""

        self.log.warning("Emergency shutdown initiated.")
        await self.notify_event(Event.EMERGENCY_SHUTDOWN)

        try:
            await self.shutdown(
                park_telescopes=True,
                disable_overwatcher=True,
                retry_without_parking=True,
                show_message=False,
            )

        except Exception as err:
            self.log.error(f"Failed to shutdown: {err!r}")
            self.log.warning("Trying to just close the dome.")

            await self.enclosure.close(
                park_telescopes=True,
                force=True,
                retry_without_parking=True,
            )

    async def observe(
        self,
        n_tile_positions: int | None = None,
        adjust_focus: bool = True,
        show_progress: bool | None = None,
        disable_tile_on_error: bool = True,
        wait_if_no_tiles: bool | float = 60.0,
    ):
        """Runs a fully automatic science observing loop."""

        # TODO: add some exception handling for keyboard interrupts.

        self.log.info("Running the cleanup recipe.")
        await self.execute_recipe("cleanup")

        self.log.info("Starting the observe loop.")

        n_completed = 0
        while True:
            try:
                tile = await run_in_executor(Tile.from_scheduler)
                break_from_while: bool = False

                for ipos, dpos in enumerate(tile.dither_positions):
                    is_last = ipos == len(tile.dither_positions) - 1
                    result, _ = await self.observer.observe_tile(
                        tile=tile,
                        dither_position=dpos,
                        async_readout=True,
                        keep_guiding=not is_last,
                        skip_slew_when_acquired=True,
                        run_cleanup=False,
                        cleanup_on_interrupt=True,
                        adjust_focus=adjust_focus,
                        show_progress=show_progress,
                    )

                    if result is False:
                        break_from_while = True
                        break

                    n_completed += 1
                    if n_tile_positions is not None and n_completed >= n_tile_positions:
                        self.log.info(
                            "Number of tile positions reached. "
                            "Finishing the observe loop."
                        )
                        break_from_while = True
                        break

                if break_from_while:
                    break

            except GortError as ee:
                if not disable_tile_on_error:
                    raise

                if ee.error_code == ErrorCode.ACQUISITION_FAILED:
                    tile_id: int | None = ee.payload.get("tile_id", None)
                    if tile_id is None:
                        self.log.error(
                            'Cannot disable tile without a "tile_id. '
                            "Continuing observations without disabling tile."
                        )
                        continue

                    await set_tile_status(tile_id, enabled=False)
                    self.log.warning(
                        f"tile_id={tile_id} has been disabled. "
                        "Continuing observations."
                    )

                elif ee.error_code == ErrorCode.SCHEDULER_CANNOT_FIND_TILE:
                    if wait_if_no_tiles is False:
                        raise

                    if wait_if_no_tiles is True:
                        wait_if_no_tiles = 60.0
                    else:
                        wait_if_no_tiles = float(wait_if_no_tiles)

                    self.log.warning(
                        "The scheduler was not able to find a valid tile to "
                        f"observe. Waiting {wait_if_no_tiles} seconds before "
                        "trying again."
                    )
                    await asyncio.sleep(wait_if_no_tiles)
                    continue

                else:
                    self.log.info(
                        "Error found in observe loop. "
                        "Running cleanup before raising the exception."
                    )
                    await self.execute_recipe("cleanup")
                    raise

            else:
                n_completed += 1

    async def run_script(self, script: str):
        """Runs a script."""

        if not script.endswith(".py"):
            script += ".py"

        path = pathlib.Path(__file__).parent / "../../scripts" / script

        cmd = await asyncio.create_subprocess_shell(f"python {path!s}")
        await cmd.communicate()

    def run_script_sync(self, script: str):
        """Runs a script."""

        if not script.endswith(".py"):
            script += ".py"

        self.log.info(f"Running script {script!r}.")

        path = pathlib.Path(__file__).parent / "../../scripts" / script
        subprocess.run(f"python {path!s}", shell=True)

    async def execute_recipe(self, recipe: str, **kwargs):
        """Executes a recipe.

        Parameters
        ----------
        recipe
            The name of the recipe to execute.
        kwargs
            Arguments to be passed to the recipe.

        """

        if recipe not in recipe_to_class:
            raise ValueError(f"Cannot find recipe {recipe!r}.")

        Recipe = recipe_to_class[recipe]

        return await Recipe(self)(**kwargs)

    async def startup(self, **kwargs):
        """Executes the :obj:`startup <.StartupRecipe>` sequence."""

        return await self.execute_recipe("startup", **kwargs)

    async def shutdown(self, **kwargs):
        """Executes the :obj:`shutdown <.ShutdownRecipe>` sequence."""

        return await self.execute_recipe("shutdown", **kwargs)

    async def cleanup(self, readout: bool = True, turn_lamps_off: bool = True):
        """Executes the :obj:`shutdown <.CleanupRecipe>` sequence."""

        return await self.execute_recipe(
            "cleanup",
            readout=readout,
            turn_lamps_off=turn_lamps_off,
        )

    async def restart_kubernetes_deployments(self, deployment: str):
        """Restarts a Kubernetes deployment."""

        self.log.warning(f"Restarting deployment {deployment!r}.")
        await kubernetes_restart_deployment(deployment)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-11-27
# @Filename: core.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Generic,
    Sequence,
    Type,
    TypeVar,
)

from packaging.version import Version

from clu.client import AMQPReply

from gort.exceptions import GortError
from gort.tools import kubernetes_list_deployments, kubernetes_restart_deployment


if TYPE_CHECKING:
    from gort.gort import Gort


GortDeviceType = TypeVar("GortDeviceType", bound="GortDevice")


class GortDeviceSet(dict[str, GortDeviceType], Generic[GortDeviceType]):
    """A set to gort-managed devices.

    Devices can be accessed as items of the :obj:`.GortDeviceSet` dictionary
    or using dot notation, as attributes.

    Parameters
    ----------
    gort
        The :obj:`.Gort` instance.
    data
        A mapping of device to device info. Each device must at least include
        an ``actor`` key with the actor to use to communicated with the device.
        Any other information is passed to the :obj:`.GortDevice` on instantiation.
    kwargs
        Other keyword arguments to pass wo the device class.

    """

    __DEVICE_CLASS__: ClassVar[Type["GortDevice"]]
    __DEPLOYMENTS__: ClassVar[list[str]] = []

    def __init__(self, gort: Gort, data: dict[str, dict], **kwargs):
        self.gort = gort

        _dict_data = {}
        for device_name in data:
            device_data = data[device_name].copy()
            actor_name = device_data.pop("actor")
            _dict_data[device_name] = self.__DEVICE_CLASS__(
                gort,
                device_name,
                actor_name,
                **device_data,
                **kwargs,
            )

        dict.__init__(self, _dict_data)

    async def init(self):
        """Runs asynchronous tasks that must be executed on init."""

        # Run devices init methods.
        results = await asyncio.gather(
            *[dev.init() for dev in self.values()],
            return_exceptions=True,
        )

        for idev, result in enumerate(results):
            if isinstance(result, Exception):
                self.write_to_log(
                    f"Failed initialising device {list(self)[idev]} "
                    f"with error {str(result)}",
                    "error",
                )

        return

    def __getattribute__(self, __name: str) -> Any:
        if __name in self:
            return self.__getitem__(__name)
        return super().__getattribute__(__name)

    async def call_device_method(self, method: Callable, *args, **kwargs):
        """Calls a method in each one of the devices.

        Parameters
        ----------
        method
            The method to call. This must be the abstract class method,
            not the method from an instantiated object.
        args,kwargs
            Arguments to pass to the method.

        """

        if not callable(method):
            raise GortError("Method is not callable.")

        if hasattr(method, "__self__"):
            # This is a bound method, so let's get the class method.
            method = method.__func__

        if not hasattr(self.__DEVICE_CLASS__, method.__name__):
            raise GortError("Method does not belong to this class devices.")

        devices = self.values()

        return await asyncio.gather(*[method(dev, *args, **kwargs) for dev in devices])

    async def send_command_all(
        self,
        command: str,
        *args,
        devices: Sequence[str] | None = None,
        **kwargs,
    ):
        """Sends a command to all the devices.

        Parameters
        ----------
        command
            The command to call.
        args, kwargs
            Arguments to pass to the :obj:`.RemoteCommand`.

        """

        tasks = []
        for name, dev in self.items():
            if devices is not None and name not in devices:
                continue

            actor_command = dev.actor.commands[command]
            tasks.append(actor_command(*args, **kwargs))

        return await asyncio.gather(*tasks)

    def write_to_log(
        self,
        message: str,
        level: str = "debug",
        header: str | None = None,
    ):
        """Writes a message to the log with a custom header.

        Parameters
        ----------
        message
            The message to log.
        level
            The level to use for logging: ``'debug'``, ``'info'``, ``'warning'``, or
            ``'error'``.
        header
            The header to prepend to the message. By default uses the class name.

        """

        if header is None:
            header = f"({self.__class__.__name__}) "

        message = f"{header}{message}"

        level = logging.getLevelName(level.upper())
        assert isinstance(level, int)

        self.gort.log.log(level, message)

    async def restart(self):
        """Restarts the set deployments and resets all controllers.

        Returns
        -------
        result
            A boolean indicting if the entire restart procedure succeeded.

        """

        failed: bool = False

        self.write_to_log("Restarting Kubernetes deployments.", "info")
        for deployment in self.__DEPLOYMENTS__:
            await kubernetes_restart_deployment(deployment)

        self.write_to_log("Waiting 15 seconds for deployments to be ready.", "info")
        await asyncio.sleep(15)

        # Check that deployments are running.
        running_deployments = await kubernetes_list_deployments()
        for deployment in self.__DEPLOYMENTS__:
            if deployment not in running_deployments:
                failed = True
                self.write_to_log(f"Deployment {deployment} did not restart.", "error")

        # Refresh the command models for all the actors.
        await asyncio.gather(*[actor.refresh() for actor in self.gort.actors.values()])

        # Refresh the device set.
        await self.init()

        return not failed


class GortDevice:
    """A gort-managed device.

    Parameters
    ----------
    gort
        The :obj:`.Gort` instance.
    name
        The name of the device.
    actor
        The name of the actor used to interface with this device. The actor is
        added to the list of :obj:`.RemoteActor` in the :obj:`.Gort`.


    """

    def __init__(self, gort: Gort, name: str, actor: str):
        self.gort = gort
        self.name = name
        self.actor = gort.add_actor(actor, device=self)

        # Placeholder version. The real one is retrieved on init.
        self.version = Version("0.99.0")

    async def init(self):
        """Runs asynchronous tasks that must be executed on init.

        If the device is part of a :obj:`.DeviceSet`, this method is called
        by :obj:`.DeviceSet.init`.

        """

        # Get the version of the actor.
        if "version" in self.actor.commands:
            try:
                reply = await self.actor.commands.version()
                if (version := reply.get("version")) is not None:
                    self.version = Version(version)
            except Exception:
                pass

        return

    def write_to_log(
        self,
        message: str,
        level: str = "debug",
        header: str | None = None,
    ):
        """Writes a message to the log with a custom header.

        Parameters
        ----------
        message
            The message to log.
        level
            The level to use for logging: ``'debug'``, ``'info'``, ``'warning'``, or
            ``'error'``.
        header
            The header to prepend to the message. By default uses the device name.

        """

        if header is None:
            header = f"({self.name}) "

        message = f"{header}{message}"

        level = logging.getLevelName(level.upper())
        assert isinstance(level, int)

        self.gort.log.log(level, message)

    def log_replies(self, reply: AMQPReply, skip_debug: bool = True):
        """Outputs command replies."""

        if reply.body:
            if reply.message_code in ["w"]:
                level = "warning"
            elif reply.message_code in ["e", "f", "!"]:
                level = "error"
            else:
                level = "debug"
                if skip_debug:
                    return

            self.write_to_log(str(reply.body), level)



# encoding: utf-8

from astropy.utils import iers
from astropy.utils.iers import conf

from sdsstools import get_config, get_package_version


# pip package name
NAME = "lvmgort"

# Loads config. config name is the package name.
config = get_config(NAME, config_envvar="GORT_CONFIG_FILE")

# package name should be pip package name
__version__ = get_package_version(path=__file__, package_name=NAME)

# Ensure that astropy does not connect to the internet.
conf.auto_download = False
conf.iers_degraded_accuracy = "ignore"

# See https://github.com/astropy/astropy/issues/15881
iers_a = iers.IERS_A.open(iers.IERS_A_FILE)
iers.earth_orientation_table.set(iers_a)

from .devices import *
from .exposure import *
from .gort import *
from .observer import *
from .remote import *
from .tile import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-08-21
# @Filename: pubsub.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import time
import uuid
import warnings

from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    ClassVar,
    Literal,
    Self,
    overload,
)

import aio_pika
from aio_pika.abc import AbstractIncomingMessage
from aio_pika.exceptions import AMQPConnectionError
from pydantic import BaseModel, Field

from gort import config
from gort.enums import Event


if TYPE_CHECKING:
    from aio_pika.abc import (
        AbstractChannel,
        AbstractConnection,
        AbstractExchange,
        AbstractQueue,
        ConsumerTag,
    )


SubCallbackType = Callable[["GortMessage"], Awaitable[None]]
MessageType = Literal["event", "notification", "custom"]


class GortPublishedMessageModel(BaseModel):
    """A model for messages published to the exchange."""

    message_type: MessageType
    event_name: int | str | None
    payload: dict[str, Any] = {}
    timestamp: float = Field(default_factory=time.time)


class GortEventModel(GortPublishedMessageModel):
    """A model for event messages."""

    event_name: str
    message_type: Literal["event"] = "event"


class GortMessage:
    """A model for messages to be published to the exchange."""

    def __init__(self, message: AbstractIncomingMessage):
        self.message = message

        self.body: dict[str, Any] = json.loads(message.body)
        self.payload: dict[str, Any] = self.body.get("payload", {})

        self.message_type: MessageType = self.body.get("message_type", "custom")

        self.event: Event | None = None
        self.event_name: str | None = None

        if self.message_type == "event":
            self.event = Event(self.body["event_name"].upper())
            self.event_name = self.event.name.upper()


def callback_wrapper(func: SubCallbackType):
    """Wraps a callback to receive a ``GortMessage`` instance."""

    async def wrapper(message: AbstractIncomingMessage):
        async with message.process():
            await func(GortMessage(message))

    return wrapper


class BasePubSub:
    """A base class to connect to a RabbitMQ GORT exchange.

    Parameters
    ----------
    connection_string
        The connection string to the RabbitMQ server.
    exchange_name
        The name of the exchange where the messages will be sent.

    """

    def __init__(
        self,
        connection_string: str | None = None,
        exchange_name: str | None = None,
    ):
        psc = config["services.pubsub"]

        self.connection_string = connection_string or psc["connection_string"]
        self.exchange_name = exchange_name or psc["exchange_name"]

        self.connection: AbstractConnection | None = None
        self.channel: AbstractChannel | None = None
        self.exchange: AbstractExchange | None = None

    async def connect(self) -> Self:
        """Connects to the RabbitMQ server and declares the exchange."""

        self.connection = await aio_pika.connect_robust(self.connection_string)

        self.channel = await self.connection.channel()
        await self.channel.set_qos(prefetch_count=1)

        self.exchange = await self.channel.declare_exchange(
            self.exchange_name,
            auto_delete=True,
            type=aio_pika.ExchangeType.FANOUT,
        )

        return self

    async def disconnect(self):
        """Disconnects from the RabbitMQ server."""

        if self.channel and not self.channel.is_closed:
            await self.channel.close()

        if self.connection and not self.connection.is_closed:
            await self.connection.close()

    async def __aenter__(self):
        if (
            not self.connection
            or self.connection.is_closed
            or not self.channel
            or self.channel.is_closed
        ):
            await self.connect()

        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.disconnect()


class GortPublisher(BasePubSub):
    """A class to publish messages to a RabbitMQ GORT exchange. A singleton."""

    _instance: ClassVar[GortPublisher]

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super(GortPublisher, cls).__new__(cls)
        return cls._instance

    def __init__(
        self,
        connection_string: str | None = None,
        exchange_name: str | None = None,
    ):
        if not hasattr(self, "connection"):
            super().__init__(connection_string, exchange_name)

    async def publish(self, message: dict, routing_key: str | None = None):
        """Publishes a message to the exchange.

        Parameters
        ----------
        message
            The message to publish. Must be a dictionary that will be encoded
            as a JSON string.
        routing_key
            The routing key to use. If not provided, uses the default routing
            key defined in the configuration.

        """

        # Give the event loop a chance to run. This should only matter if the
        # library is being used in IPython, which does run the loop unless there
        # is a command running.
        try:
            await asyncio.sleep(0.1)
        except Exception:
            pass
        finally:
            if not self.channel or not self.exchange or self.channel.is_closed:
                await self.connect()

        assert self.exchange, "exchange not defined."

        for _ in range(3):
            try:
                await self.exchange.publish(
                    aio_pika.Message(body=json.dumps(message).encode()),
                    routing_key=routing_key or config["services.pubsub.routing_key"],
                )
                return
            except AMQPConnectionError:
                await self.connect()
            except Exception as err:
                from gort.exceptions import GortWarning

                warnings.warn(f"Unexpected error in GortPublisher: {err}", GortWarning)
                break


class GortSubscriber(BasePubSub):
    """A class to subscribe to messages from a RabbitMQ GORT exchange."""

    def __init__(
        self,
        connection_string: str | None = None,
        exchange_name: str | None = None,
        callback: SubCallbackType | None = None,
        queue_name: str | None = None,
    ):
        super().__init__(
            connection_string=connection_string,
            exchange_name=exchange_name,
        )

        self.queue_name: str | None = queue_name
        self.queue: AbstractQueue | None = None
        self.callback = callback

        self.consumer_tag: ConsumerTag | None = None

    async def connect(self, queue_name: str | None = None) -> Self:
        """Connects to the exchange, declares a queue, and binds the callback.

        Parameters
        ----------
        queue_name
            The name of the queue to declare. If not provided, a random name
            will be generated (recommended).

        """

        await super().connect()

        assert self.channel, "channel not defined."
        assert self.exchange, "exchange not defined."

        self.queue_name = (
            queue_name
            or self.queue_name
            or f"{self.exchange_name}-{str(uuid.uuid4()).split('-')[-1]}"
        )

        self.queue = await self.channel.declare_queue(
            self.queue_name,
            auto_delete=True,
            exclusive=True,
        )

        await self.queue.bind(self.exchange, routing_key="data")

        if self.callback:
            await self.queue.consume(callback_wrapper(self.callback))

        return self

    async def disconnect(self):
        """Disconnects from the RabbitMQ server."""

        if self.callback and self.queue and self.consumer_tag:
            await self.queue.cancel(self.consumer_tag)

            if self.exchange:
                await self.queue.unbind(self.exchange)

        await super().disconnect()

    @overload
    async def iterator(
        self,
        decode: Literal[True] = True,
    ) -> AsyncGenerator[GortMessage, None]: ...

    @overload
    async def iterator(
        self,
        decode: Literal[False],
    ) -> AsyncGenerator[AbstractIncomingMessage, None]: ...

    async def iterator(
        self,
        decode: bool = True,
    ) -> AsyncGenerator[AbstractIncomingMessage | GortMessage, None]:
        """Iterates over a queue and yields messages."""

        async with self as instance:
            assert instance.queue, "queue not defined."

            async with instance.queue.iterator() as queue_iter:
                async for message in queue_iter:
                    async with message.process():
                        if decode:
                            yield GortMessage(message)
                        else:
                            yield message


async def notify_event(event: Event, payload: dict[str, Any] = {}):
    """Notifies an event to the exchange."""

    message = GortEventModel(event_name=event.name, payload=payload).model_dump()
    await GortPublisher().publish(message)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-08
# @Filename: tile.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import warnings

from typing import Sequence, cast

import polars
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time
from httpx import RequestError

from gort import config
from gort.exceptions import (
    ErrorCode,
    GortError,
    GortNotImplemented,
    GortWarning,
    TileError,
)
from gort.tools import (
    get_by_source_id,
    get_calibrators_sync,
    get_db_connection,
    get_next_tile_id_sync,
    set_tile_status,
)
from gort.transforms import fibre_to_master_frame, offset_to_master_frame_pixel


__all__ = [
    "Coordinates",
    "QuerableCoordinates",
    "ScienceCoordinates",
    "SkyCoordinates",
    "StandardCoordinates",
    "Tile",
]


CoordTuple = tuple[float, float]


class Coordinates:
    """Basic coordinates class.

    Parameters
    ----------
    ra
        The RA coordinate, in degrees. FK5 frame at the epoch of observation.
    dec
        The Dec coordinate, in degrees.
    pa
        Position angle of the IFU. Defaults to PA=0.
    centre_on_fibre
        The name of the fibre on which to centre the target, with the format
        ``<ifulabel>-<finufu>``. By default, acquires the target on the central
        fibre of the science IFU.

    """

    def __init__(
        self,
        ra: float,
        dec: float,
        pa: float | None = None,
        centre_on_fibre: str | None = None,
    ):
        self.ra = ra
        self.dec = dec
        self.pa = pa % 360 if pa is not None else 0.0

        self.skycoord = SkyCoord(ra=ra, dec=dec, unit="deg", frame="fk5")

        self.centre_on_fibre = centre_on_fibre

        # The MF pixel on which to guide/centre the target.
        self._mf_pixel = self.set_mf_pixel(centre_on_fibre)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} "
            f"(ra={self.ra:.6f}, dec={self.dec:.6f}, pa={self.pa:.3f})>"
        )

    def __str__(self):
        return f"{self.ra:.6f}, {self.dec:.6f}, {self.pa:.3f}"

    def calculate_altitude(self, time: Time | None = None):
        """Returns the current altitude of the target."""

        if time is None:
            time = Time.now()

        location = EarthLocation.from_geodetic(**config["site"])

        sc = self.skycoord.copy()
        sc.obstime = time
        sc.location = location
        altaz = sc.transform_to("altaz")

        return altaz.alt.deg

    def is_observable(self):
        """Determines whether a target is observable."""

        return self.calculate_altitude() > 30

    def set_mf_pixel(self, fibre_name: str | None = None, xz: CoordTuple | None = None):
        """Calculates and sets the master frame pixel on which to centre the target.

        If neither ``fibre_name`` or ``xz`` are passed, resets to centring
        the target on the central fibre of the IFU.

        Parameters
        ----------
        fibre_name
            The fibre to which to centre the target, with the format
            ``<ifulabel>-<finifu>``.
        xz
            The coordinates, in master frame pixels, on which to centre
            the target.

        Returns
        -------
        pixel
            A tuple with the x and z coordinates of the pixel in the master frame,
            or :obj:`None` if resetting to the central fibre.

        """

        if fibre_name is not None:
            xmf, zmf = fibre_to_master_frame(fibre_name)
        elif xz is not None:
            xmf, zmf = xz
        else:
            self._mf_pixel = None
            return None

        self._mf_pixel = (xmf, zmf)

        return (xmf, zmf)


class QuerableCoordinates(Coordinates):
    """A class of coordinates that can be retrieved from the database."""

    __db_table__: str = ""
    targets: SkyCoord | None = None

    @classmethod
    def from_science_coordinates(
        cls,
        sci_coords: ScienceCoordinates,
        exclude_coordinates: Sequence[CoordTuple] = [],
        exclude_invisible: bool = True,
    ):
        """Retrieves a valid and observable position from the database.

        Parameters
        ----------
        sci_coords
            The science coordinates. The position selected will be the
            closest to these coordinates.
        exclude_coordinates
            A list of RA/Dec coordinates to exclude. No region closer
            than one degree to these coordinates will be selected.
        exclude_invisible
            Exclude targets that are too low.

        """

        connection = get_db_connection()

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            targets = polars.read_database(
                f"SELECT ra,dec from {cls.__db_table__};",
                connection,
            )

        # Cache query.
        if cls.targets is None:
            cls.targets = SkyCoord(
                ra=targets["ra"].to_list(),
                dec=targets["dec"].to_list(),
                unit="deg",
            )

        assert cls.targets is not None
        skycoords = cls.targets.copy()

        # Exclude regions too close to the exlcuded ones.
        for ex_coords in exclude_coordinates:
            ex_skycoords = SkyCoord(ra=ex_coords[0], dec=ex_coords[1], unit="deg")
            skycoords = skycoords[skycoords.separation(ex_skycoords).deg > 1]

        # Exclude targets that are too low.
        if exclude_invisible:
            skycoords.location = EarthLocation.from_geodetic(**config["site"])

            skycoords.obstime = Time.now()
            altaz_skycoords = skycoords.transform_to("altaz")
            skycoords = skycoords[altaz_skycoords.alt.deg > 30]

        if len(skycoords) == 0:
            raise TileError("No sky coordinates found.")

        seps = skycoords.separation(sci_coords.skycoord)
        skycoord_min = skycoords[seps.argmin()]

        return cls(skycoord_min.ra.deg, skycoord_min.dec.deg)

    def verify_and_replace(self, exclude_coordinates: Sequence[CoordTuple] = []):
        """Verifies that the coordinates are visible and if not, replaces them.

        Parameters
        ----------
        exclude_coordinates
            A list of RA/Dec coordinates to exclude. No region closer
            than one degree to these coordinates will be selected.

        """

        if not self.is_observable():
            # Use current coordinates as proxy for the science telescope.
            sci_coords = ScienceCoordinates(self.ra, self.dec)
            valid_skycoords = self.from_science_coordinates(
                sci_coords,
                exclude_coordinates=exclude_coordinates,
            )
            super().__init__(valid_skycoords.ra, valid_skycoords.dec)


class ScienceCoordinates(Coordinates):
    """A science position.

    Parameters
    ----------
    ra
        The RA coordinate, in degrees. FK5 frame at the epoch of observation.
    dec
        The Dec coordinate, in degrees.
    pa
        Position angle of the IFU. Defaults to PA=0.
    centre_on_fibre
        The name of the fibre on which to centre the target, with the format
        ``<ifulabel>-<finufu>``. By default, acquires the target on the central
        fibre of the science IFU.

    """

    dither_position: int = 0


class SkyCoordinates(QuerableCoordinates):
    """A sky position.

    In addition to the `.QuerableCoordinates` arguments the class accepts
    a ``name`` identifier.

    """

    __db_table__ = "lvmopsdb.sky"

    def __init__(self, *args, name: str | None = None, **kwargs):
        self.name = name
        self.pk: int | None = None

        super().__init__(*args, **kwargs)


class StandardCoordinates(QuerableCoordinates):
    """A standard position.

    In addition to the `.QuerableCoordinates` arguments the class accepts
    a ``source_id`` Gaia identifier.

    """

    __db_table__ = "lvmopsdb.standard"

    def __init__(
        self,
        ra: float | None = None,
        dec: float | None = None,
        source_id: int | None = None,
        **kwargs,
    ):
        self.source_id = source_id
        self.pk: int | None = None

        if ra is None or dec is None:
            if source_id is None:
                raise TileError("Must pass either ra/dec or source_id.")
            else:
                if (data := get_by_source_id(int(source_id))) is None:
                    raise TileError(f"Cannot find Gaia data for source_id={source_id}.")

                ra = data["ra"]
                dec = data["dec"]

        assert ra is not None and dec is not None

        super().__init__(ra, dec, **kwargs)


SpecCoordsType = Sequence[StandardCoordinates | CoordTuple | int | dict] | None
SkyCoordsType = dict[str, SkyCoordinates] | dict[str, CoordTuple] | None


class Tile(dict[str, Coordinates | Sequence[Coordinates] | None]):
    """A representation of a science pointing with associated calibrators.

    This class is most usually initialised from a classmethod like
    :obj:`.from_scheduler`.

    Parameters
    ----------
    sci_coords
        The science telescope pointing.
    sky_coords
        A dictionary of ``skye`` and ``skyw`` coordinates.
    spec_coords
        A list of coordinates to observe with the spectrophotometric telescope.
    dither_position
        The dither position(s) to obseve.
    object
        The name of the object.
    allow_replacement
        If :obj:`True`, allows the replacement of empty, invalid or low altitude
        sky and standard targets.

    """

    def __init__(
        self,
        sci_coords: ScienceCoordinates,
        sky_coords: SkyCoordsType = None,
        spec_coords: SpecCoordsType = None,
        dither_positions: int | Sequence[int] = 0,
        object: str | None = None,
        allow_replacement: bool = True,
    ):
        self.allow_replacement = allow_replacement

        self.tile_id: int | None = None
        self.dither_positions = (
            dither_positions
            if isinstance(dither_positions, Sequence)
            else [dither_positions]
        )

        self.object = object or (f"Tile {self.tile_id}" if self.tile_id else None)

        dict.__init__(self, {})

        self.set_sci_coords(sci_coords)
        if isinstance(dither_positions, int):
            self.set_dither_position(dither_positions)
        else:
            self.set_dither_position(dither_positions[0])

        self.set_sky_coords(sky_coords, allow_replacement=allow_replacement)
        self.set_spec_coords(spec_coords, reject_invisible=allow_replacement)

    def __repr__(self):
        return (
            "<Tile "
            f"(tile_id={self.tile_id}, "
            f"science ra={self.sci_coords.ra:.6f}, "
            f"dec={self.sci_coords.dec:.6f}, "
            f"pa={self.sci_coords.pa:.3f}; "
            f"n_skies={len(self.sky_coords)}; "
            f"n_standards={len(self.spec_coords)}; "
            f"dither_positions={self.dither_positions!r})>"
        )

    def set_dither_position(self, dither: int):
        """Sets the full frame pixel for the science IFU to the dither position."""

        raoff: float
        decoff: float
        raoff, decoff = config["guiders"]["devices"]["sci"]["dither_offsets"][dither]

        xx, zz = offset_to_master_frame_pixel(ra=raoff, dec=decoff)
        self.sci_coords.set_mf_pixel(xz=(xx, zz))

        self.sci_coords.dither_position = dither

    @property
    def sci_coords(self):
        """Returns the science coordinates."""

        return cast(ScienceCoordinates, self["sci"])

    @sci_coords.setter
    def sci_coords(self, new_coords: ScienceCoordinates):
        """Sets the science coordinates."""

        if isinstance(new_coords, (list, tuple)):
            new_coords = ScienceCoordinates(*new_coords)

        self["sci"] = new_coords

    @property
    def sky_coords(self) -> dict[str, SkyCoordinates]:
        """Returns the sky coordinates."""

        skyw = cast(SkyCoordinates, self["skyw"])
        skye = cast(SkyCoordinates, self["skye"])

        return {"skye": skye, "skyw": skyw}

    @sky_coords.setter
    def sky_coords(self, new_coords: dict[str, SkyCoordinates]):
        """Returns the sky coordinates."""

        for tel in ["skye", "skyw"]:
            sky = new_coords.get(tel, None)
            if isinstance(sky, (tuple, list)):
                sky = SkyCoordinates(*sky)
            self[tel] = sky

    @property
    def spec_coords(self):
        """Returns the Spec coordinates."""

        return cast(Sequence[StandardCoordinates], self["spec"])

    @spec_coords.setter
    def spec_coords(self, new_coords: Sequence[StandardCoordinates]):
        """Sets the SkyW coordinates."""

        parsed_coords: Sequence[Coordinates] = []
        for coords in new_coords:
            if isinstance(coords, (list, tuple)):
                parsed_coords.append(StandardCoordinates(*coords))
            else:
                parsed_coords.append(coords)

        self["spec"] = parsed_coords

    @classmethod
    def from_coordinates(
        cls,
        ra: float,
        dec: float,
        pa: float = 0.0,
        sky_coords: SkyCoordsType = None,
        spec_coords: SpecCoordsType | None = None,
        **kwargs,
    ):
        """Creates an instance from coordinates, allowing autocompletion.

        Parameters
        ----------
        ra,dec
            The science telescope pointing.
        pa
            Position angle of the science IFU. Defaults to PA=0.
        sky_coords
            A dictionary of ``skye`` and ``skyw`` coordinates. If `None`,
            autocompleted from the closest available regions.
        spec_coords
            A list of coordinates to observe with the spectrophotometric telescope.
            If :obj:`None`, selects the 12 closest standard stars.
        kwargs
            Arguments to be passed to the initialiser.

        """

        sci_coords = ScienceCoordinates(ra, dec, pa=pa)

        calibrators: dict | None = None
        if sky_coords is None or spec_coords is None:
            calibrators = get_calibrators_sync(ra=ra, dec=dec)

        if sky_coords is None:
            assert calibrators is not None
            sky_coords = {}
            sky_coords["skye"] = SkyCoordinates(*calibrators["sky_pos"][0])
            sky_coords["skyw"] = SkyCoordinates(*calibrators["sky_pos"][1])

            sky_coords["skye"].pk = calibrators["sky_pks"][0]
            sky_coords["skyw"].pk = calibrators["sky_pks"][1]

        if spec_coords is None:
            assert calibrators is not None
            spec_coords = []
            for ii in range(12):
                coords = StandardCoordinates(
                    *calibrators["standard_pos"][ii],
                    source_id=calibrators["standard_ids"][ii],
                )
                coords.pk = calibrators["standard_pks"][ii]
                spec_coords.append(coords)

        return cls(
            sci_coords,
            spec_coords=spec_coords,
            sky_coords=sky_coords,
            allow_replacement=False,
            **kwargs,
        )

    @classmethod
    def from_scheduler(
        cls,
        tile_id: int | None = None,
        ra: float | None = None,
        dec: float | None = None,
        pa: float = 0.0,
        **kwargs,
    ):
        """Creates a new instance of :obj:`.Tile` with data from the scheduler.

        Parameters
        ----------
        tile_id
            The ``tile_id`` for which to create a new :obj:`.Tile`. If :obj:`None`,
            and ``ra`` and ``dec`` are also :obj:`None`, the best ``tile_id``
            selected by the scheduler will be used.
        ra
            Right ascension coordinates of the science telescope pointing.
            Calibrators will be selected from the scheduler.
        dec
            Declination coordinates of the science telescope pointing.
        pa
            Position angle of the science IFU. Defaults to PA=0.
        kwargs
            Arguments to be passed to the initialiser.

        """

        if tile_id is None and ra is None and dec is None:
            try:
                tile_id_data = get_next_tile_id_sync()
            except RequestError:
                raise TileError(
                    "Cannot retrieve tile_id from scheduler.",
                    error_code=ErrorCode.SCHEDULER_CANNOT_FIND_TILE,
                )

            tile_id = tile_id_data["tile_id"]
            sci_pos = tile_id_data["tile_pos"]
            dither_pos = tile_id_data["dither_pos"]

            if tile_id is None or tile_id < 0:
                raise GortError(
                    "The scheduler could not find a valid tile to observe.",
                    error_code=ErrorCode.SCHEDULER_CANNOT_FIND_TILE,
                )

        elif tile_id is not None:
            raise GortNotImplemented("Initialising from a tile_id is not supported.")

        elif not tile_id and (ra is not None and dec is not None):
            tile_id = None
            sci_pos = (ra, dec, pa)
            dither_pos = 0

        else:
            raise TileError("Invalid inputs.")

        sci_coords = ScienceCoordinates(*sci_pos, centre_on_fibre=None)

        if tile_id:
            calibrator_data = get_calibrators_sync(tile_id=tile_id)
        else:
            calibrator_data = get_calibrators_sync(ra=sci_pos[0], dec=sci_pos[1])

        sky_coords = {
            "skye": SkyCoordinates(
                *calibrator_data["sky_pos"][0],
                name=calibrator_data["sky_names"][0],
            ),
            "skyw": SkyCoordinates(
                *calibrator_data["sky_pos"][1],
                name=calibrator_data["sky_names"][1],
            ),
        }

        sky_coords["skye"].pk = calibrator_data["sky_pks"][0]
        sky_coords["skyw"].pk = calibrator_data["sky_pks"][1]

        spec_coords = []
        for ii in range(len(calibrator_data["standard_pos"])):
            std_coords = StandardCoordinates(
                *calibrator_data["standard_pos"][ii],
                source_id=calibrator_data["standard_ids"][ii],
            )
            std_coords.pk = calibrator_data["standard_pks"][ii]
            spec_coords.append(std_coords)

        new_obj = cls(
            sci_coords,
            sky_coords=sky_coords,
            spec_coords=spec_coords,
            dither_positions=dither_pos,
            **kwargs,
        )
        new_obj.tile_id = tile_id

        return new_obj

    def set_sci_coords(
        self,
        sci_coords: ScienceCoordinates | CoordTuple,
    ) -> ScienceCoordinates:
        """Sets the science telescope coordinates.

        Parameters
        ----------
        sci_coords
            A :obj:`.ScienceCoordinates` object or a tuple with RA/Dec
            coordinates for the science telescope.

        """

        if isinstance(sci_coords, ScienceCoordinates):
            self.sci_coords = sci_coords
        else:
            self.sci_coords = ScienceCoordinates(*sci_coords)

        return self.sci_coords

    def set_sky_coords(
        self,
        sky_coords: SkyCoordsType = None,
        allow_replacement: bool = True,
    ) -> dict[str, SkyCoordinates]:
        """Sets the sky telescopes coordinates.

        Parameters
        ----------
        sky_coords
            A dictionary of ``skye`` and ``skyw`` coordinates. Each value must
            be a :obj:`.SkyCoordinates` object or a tuple of RA/Dec coordinates.
        allow_replacement
            If :obj:`True`, allows the replacement of empty, invalid or low
            altitude targets.

        """

        if sky_coords is None:
            sky_coords = {}

        valid_sky_coords: dict[str, SkyCoordinates] = {}
        assigned_coordinates: Sequence[CoordTuple] = []

        for telescope in ["skye", "skyw"]:
            tel_coords = sky_coords.get(telescope, None)

            replace: bool = False
            if tel_coords is None:
                replace = True
            elif isinstance(tel_coords, SkyCoordinates):
                tel_coords = tel_coords
            else:
                tel_coords = SkyCoordinates(*tel_coords)

            if allow_replacement is False:
                if tel_coords is not None:
                    valid_sky_coords[telescope] = tel_coords
                continue

            # If both coordinates are assigned, check that they are not identical.
            if (
                tel_coords is not None
                and telescope == "skyw"
                and "skye" in valid_sky_coords
            ):
                if (
                    tel_coords.ra == valid_sky_coords["skye"].ra
                    and tel_coords.dec == valid_sky_coords["skye"].dec
                ):
                    tel_coords = None
                    replace = True

            if replace:
                try:
                    tel_coords = SkyCoordinates.from_science_coordinates(
                        self.sci_coords,
                        exclude_coordinates=assigned_coordinates,
                    )
                except Exception as err:
                    warnings.warn(
                        f"Failed getting sky coordinates for {telescope}: {err}",
                        GortWarning,
                    )
                    continue

            try:
                assert tel_coords is not None
                tel_coords.verify_and_replace(
                    exclude_coordinates=assigned_coordinates,
                )
            except Exception as err:
                warnings.warn(
                    f"Failed verifying sky coordinates for {telescope}: {err}",
                    GortWarning,
                )
                continue

            assigned_coordinates.append((tel_coords.ra, tel_coords.dec))
            valid_sky_coords[telescope] = tel_coords

        self.sky_coords = valid_sky_coords

        return self.sky_coords

    def set_spec_coords(
        self,
        spec_coords: SpecCoordsType = None,
        reject_invisible: bool = True,
    ) -> Sequence[StandardCoordinates]:
        """Sets the spec telescope coordinates.

        Parameters
        ----------
        spec_coords
            A list of coordinates to observe with the spectrophotometric telescope.
        reject_invisible
            Skip targets that are not visible now.

        """

        valid_spec_coords = []

        if spec_coords is None:
            pass
        else:
            for coords in spec_coords:
                if isinstance(coords, Coordinates):
                    pass
                elif isinstance(coords, (list, tuple)):
                    coords = StandardCoordinates(*coords)
                elif isinstance(coords, int):
                    coords = StandardCoordinates(source_id=coords)
                elif isinstance(coords, dict):
                    coords = StandardCoordinates(**coords)
                else:
                    raise TypeError(f"Invalid spec coordinate {coords!r}.")

                if reject_invisible and not coords.is_observable():
                    continue

                valid_spec_coords.append(coords)

        self.spec_coords = valid_spec_coords

        return self.spec_coords

    async def enable(self):
        """Enables the tile for observation."""

        if self.tile_id is None:
            raise TileError("Cannot enable tile without a tile_id.")

        await set_tile_status(self.tile_id, enabled=True)

    async def disable(self):
        """Disables the tile for observation."""

        if self.tile_id is None:
            raise TileError("Cannot disable tile without a tile_id.")

        await set_tile_status(self.tile_id, enabled=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-26
# @Filename: weather.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import enum
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from time import time

from typing import TYPE_CHECKING, cast

import polars

from gort.overwatcher.core import OverwatcherModuleTask
from gort.overwatcher.overwatcher import OverwatcherModule
from gort.tools import get_lvmapi_route


if TYPE_CHECKING:
    pass


__all__ = ["WeatherOverwatcher"]


class WeatherRisk(enum.Enum):
    """An enumeration of weather risk levels."""

    SAFE = enum.auto()
    CAUTION = enum.auto()
    DANGER = enum.auto()
    EXTREME = enum.auto()
    UNKNOWN = enum.auto()


@dataclass
class WeatherState:
    """A dataclass to store the weather state."""

    temperature_10: float
    wind_speed_10: float
    wind_speed_30: float
    wind_speed_10: float
    wind_speed_30: float
    gust_10: float
    rh_10: float
    rh_30: float
    lvm_rain_sensor_alarm: bool
    rain_intensity_10: float
    station: str


class WeatherMonitorTask(OverwatcherModuleTask["WeatherOverwatcher"]):
    """Monitors the weather state."""

    name = "weather_monitor"
    keep_alive = True
    restart_on_error = True

    async def task(self):
        """Updates the weather data."""

        n_failures: int = 0

        while True:
            try:
                await self.update_weather()
                await self.module.check_weather()
            except Exception as err:
                if self.unavailable is False:
                    self.log.error(f"Failed to get weather data: {err!r}")
                n_failures += 1
            else:
                self.last_updated = time()
                self.unavailable = False
                n_failures = 0
            finally:
                if self.unavailable is False and n_failures >= 5:
                    self.unavailable = True
                    self.log.critical(
                        "Failed to get weather data 5 times. "
                        "Triggering an emergency shutdown.",
                    )
                    asyncio.create_task(
                        self.overwatcher.shutdown(
                            reason="weather data unavailable",
                            park=False,
                        )
                    )

            await asyncio.sleep(60)

    async def update_weather(self):
        """Processes the weather update and determines whether it is safe to observe."""

        self.data = await self.module.get_weather_report()
        is_raining = await self.module.is_raining()

        if self.data is None or len(self.data) == 0:
            raise ValueError("No weather data available.")

        # Check that the data is recent.
        now = datetime.now(timezone.utc)
        last_point = cast(datetime, self.data["ts"].max())
        last_point_delay = (now - last_point).total_seconds()

        if last_point_delay > 300:
            raise ValueError("Weather data is too old.")

        # Data for the last 10 and 30 minutes.
        now_10 = now - timedelta(minutes=10)
        now_30 = now - timedelta(minutes=30)
        data_10 = self.data.filter(self.data["ts"] > now_10)
        data_30 = self.data.filter(self.data["ts"] > now_30)

        avg_10 = data_10.select(polars.col(polars.Float64)).mean()
        avg_30 = data_30.select(polars.col(polars.Float64)).mean()

        self.state = WeatherState(
            temperature_10=avg_10["temperature"][0],
            wind_speed_10=avg_10["wind_speed_avg"][0],
            wind_speed_30=avg_30["wind_speed_avg"][0],
            gust_10=cast(float, data_10["wind_speed_max"].max()),
            rh_10=avg_10["relative_humidity"][0],
            rh_30=avg_30["relative_humidity"][0],
            lvm_rain_sensor_alarm=is_raining,
            rain_intensity_10=cast(float, data_10["rain_intensity"].max()),
            station=self.data["station"][0],
        )

        # Determine the risk level.
        new_risk = WeatherRisk.SAFE

        if self.state.temperature_10 < -10:
            new_risk = WeatherRisk.EXTREME

        if self.state.wind_speed_10 > 50 or self.state.wind_speed_30 > 40:
            new_risk = WeatherRisk.DANGER

        if self.state.lvm_rain_sensor_alarm:
            new_risk = WeatherRisk.EXTREME

        if self.state.rh_10 > 80:
            new_risk = WeatherRisk.DANGER

        self.module.risk = new_risk


class WeatherOverwatcher(OverwatcherModule):
    """Monitors weather conditions."""

    name = "weather"

    tasks = [WeatherMonitorTask()]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.data: polars.DataFrame | None = None
        self.last_updated: float = 0.0

        self.unavailable: bool = False

        self.state: WeatherState | None = None
        self.risk = WeatherRisk.UNKNOWN

    async def check_weather(self):
        """Checks the weather state and triggers an emergency shutdown if necessary."""

        if await self.overwatcher.dome.is_closing():
            return

        if not self.is_safe():
            await self.overwatcher.shutdown(
                reason="unsafe weather conditions",
                park=True,
            )

    def is_safe(self):
        """Determines whether it is safe to open."""

        if self.risk in [WeatherRisk.DANGER, WeatherRisk.EXTREME, WeatherRisk.UNKNOWN]:
            return False

        return True

    @staticmethod
    async def get_weather_report(delta_time=3600) -> polars.DataFrame:
        """Returns a weather report."""

        data = await get_lvmapi_route("/weather/report", delta_time=delta_time)

        df = polars.DataFrame(data)
        df = df.with_columns(
            ts=polars.col("ts").str.to_datetime(
                time_unit="ms",
                time_zone="UTC",
            )
        )

        return df.sort("ts")

    @staticmethod
    async def is_raining():
        """Determines whether it is raining according to the LVM rain sensor."""

        data = await get_lvmapi_route("/enclosure")

        return "RAIN_SENSOR_ALARM" in data["safety_status"]["labels"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-26
# @Filename: overwatcher.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import dataclasses
import pathlib
from copy import copy
from time import time

from typing import cast

from sdsstools import Configuration
from sdsstools.utils import GatheringTaskGroup

from gort.exceptions import GortError
from gort.gort import Gort
from gort.overwatcher.core import OverwatcherBaseTask, OverwatcherModule
from gort.overwatcher.helpers import DomeHelper
from gort.overwatcher.helpers.notifier import NotifierMixIn
from gort.overwatcher.helpers.tasks import DailyTasks
from gort.overwatcher.troubleshooter.troubleshooter import Troubleshooter
from gort.tools import LogNamespace


@dataclasses.dataclass
class OverwatcherState:
    """Dataclass with the overwatcher state values."""

    running: bool = False
    enabled: bool = False
    observing: bool = False
    calibrating: bool = False
    troubleshooting: bool = False
    focusing: bool = False
    night: bool = False
    safe: bool = False
    allow_calibrations: bool = True
    dry_run: bool = False


class OverwatcherTask(OverwatcherBaseTask):
    """Overwatcher task that is aware of the overwatcher instance."""

    def __init__(self, overwatcher: Overwatcher):
        super().__init__()

        self.overwatcher = overwatcher

        # A bit configuring but _log is used internally, mainly for
        # OverwatcherBaseTask.run() and log is for external use.
        self._log = self.overwatcher.log
        self.log = self._log


class OverwatcherMainTask(OverwatcherTask):
    """The main overwatcher task."""

    name = "overwatcher_task"
    keep_alive = True
    restart_on_error = True

    def __init__(self, overwatcher: Overwatcher):
        super().__init__(overwatcher)

        self.previous_state = OverwatcherState()
        self._pending_close_dome: bool = False

    async def task(self):
        """Main overwatcher task."""

        await asyncio.sleep(1)

        ow = self.overwatcher

        while True:
            self.previous_state = copy(ow.state)

            try:
                is_safe, _ = ow.alerts.is_safe()
                is_night = ow.ephemeris.is_night()

                ow.state.night = is_night
                ow.state.safe = is_safe
                ow.state.observing = ow.observer.is_observing
                ow.state.focusing = ow.observer.focusing
                ow.state.troubleshooting = ow.troubleshooter.is_troubleshooting()

                running_calibration = ow.calibrations.get_running_calibration()
                ow.state.calibrating = running_calibration is not None

                # TODO: should these handlers be scheduled as tasks? Right now
                # they can block for a good while until the dome is open/closed.

                if not is_safe:
                    await self.handle_unsafe()

                if not is_night:
                    await self.handle_daytime()

                if not ow.state.enabled:
                    await self.handle_disabled()

                # Run daily tasks.
                await ow.daily_tasks.run_all()

            except Exception as err:
                await ow.notify(
                    f"Error in main overwatcher task: {err!r}",
                    level="error",
                )

                # Avoid rapid fire errors. Sleep a bit longer before trying again.
                await asyncio.sleep(30)

            await asyncio.sleep(5)

    async def handle_unsafe(self):
        """Closes the dome if the conditions are unsafe."""

        closed = await self.overwatcher.dome.is_closing()

        observing = self.overwatcher.observer.is_observing
        cancelling = self.overwatcher.observer.is_cancelling
        calibrating = self.overwatcher.state.calibrating

        # TODO: should this only happen if the overwatcher is enabled?

        if not closed or observing or calibrating:
            try:
                async with asyncio.timeout(delay=30):
                    await self.overwatcher.notify(
                        "Unsafe conditions detected.",
                        level="warning",
                    )

                    if observing and not cancelling:
                        try:
                            await self.overwatcher.observer.stop_observing(
                                immediate=True,
                                reason="unsafe conditions",
                            )
                        except Exception as err:
                            await self.overwatcher.notify(
                                f"Error stopping observing: {err!r}",
                                level="error",
                            )
                            await self.overwatcher.notify(
                                "I will close the dome anyway.",
                                level="warning",
                            )

                    if calibrating:
                        await self.overwatcher.calibrations.cancel()

            except asyncio.TimeoutError:
                await self.overwatcher.notify(
                    "Timed out while handling unsafe conditions.",
                    level="error",
                )

            except Exception as err:
                await self.overwatcher.notify(
                    f"Error handling unsafe conditions: {err!r}",
                    level="error",
                )

            finally:
                if not closed:
                    await self.overwatcher.notify("Closing the dome.")
                    await self.overwatcher.dome.shutdown(retry=True, park=True)

                    # If we have to close because of unsafe conditions, we don't want
                    # to reopen too soon. We lock the dome for 30 minutes.
                    self.overwatcher.alerts.locked_until = time() + 1800

    async def handle_daytime(self):
        """Handles daytime."""

        # Don't do anything if we are calibrating. If a calibration script opened the
        # dome it should close it afterwards.
        if self.overwatcher.state.calibrating:
            return

        # Also don't do anything if the overwatcher is not enabled.
        # TODO: maybe we should close the dome if it's open?
        if not self.overwatcher.state.enabled:
            return

        observing = self.overwatcher.observer.is_observing
        cancelling = self.overwatcher.observer.is_cancelling

        if observing and not cancelling:
            # Decide whether to complete the current exposure or stop immediately.
            exposure_finishes_at = self.overwatcher.observer.next_exposure_completes
            now = time()

            if exposure_finishes_at > 0 and (exposure_finishes_at - now) < 300:
                immediate = False
                notification_message = "Finishing this exposure and closing the dome."

            else:
                immediate = True
                notification_message = "Cancelling exposure and closing the dome."

            await self.overwatcher.notify("Twilight reached. " + notification_message)

            await self.overwatcher.observer.stop_observing(
                immediate=immediate,
                reason="daytime conditions",
            )
            self._pending_close_dome = True

        if not observing and not cancelling and self._pending_close_dome:
            try:
                closed = await self.overwatcher.dome.is_closing()
                if not closed:
                    await self.overwatcher.dome.shutdown(retry=True, park=True)
            finally:
                self._pending_close_dome = False

    async def handle_disabled(self):
        """Handles the disabled state."""

        observing = self.overwatcher.observer.is_observing
        cancelling = self.overwatcher.observer.is_cancelling

        if observing and not cancelling:
            # Disable after this tile.
            await self.overwatcher.observer.stop_observing(
                immediate=False,
                reason="overwatcher was disabled",
            )


class OverwatcherPingTask(OverwatcherTask):
    """Emits a ping notification every five minutes."""

    name = "overwatcher_ping"
    keep_alive = True
    restart_on_error = True

    delay: float = 900

    async def task(self):
        """Ping task."""

        while True:
            await asyncio.sleep(self.delay)
            self.log.debug("I am alive!")


class Overwatcher(NotifierMixIn):
    """Monitors the observatory."""

    instance: Overwatcher | None = None

    def __new__(cls, *args, **kwargs):
        if not cls.instance:
            cls.instance = super(Overwatcher, cls).__new__(cls)
        return cls.instance

    def __init__(
        self,
        gort: Gort | None = None,
        verbosity: str = "debug",
        calibrations_file: str | pathlib.Path | None = None,
        dry_run: bool = False,
        **kwargs,
    ):
        from gort.overwatcher import (
            AlertsOverwatcher,
            CalibrationsOverwatcher,
            EphemerisOverwatcher,
            EventsOverwatcher,
            ObserverOverwatcher,
            SafetyOverwatcher,
            TransparencyOverwatcher,
        )

        # Check if the instance already exists, in which case do nothing.
        if hasattr(self, "gort"):
            return

        self.gort = gort or Gort(verbosity=verbosity, **kwargs)
        self.config = cast(Configuration, self.gort.config)
        self.log = LogNamespace(self.gort.log, header=f"({self.__class__.__name__}) ")

        self.state = OverwatcherState()
        self.state.dry_run = dry_run

        self.dome = DomeHelper(self)
        self.troubleshooter = Troubleshooter(self)
        self.tasks: list[OverwatcherTask] = [
            OverwatcherMainTask(self),
            OverwatcherPingTask(self),
        ]

        # A series of tasks that must be run once every day.
        self.daily_tasks = DailyTasks(self)

        self.safety = SafetyOverwatcher(self)
        self.ephemeris = EphemerisOverwatcher(self)
        self.calibrations = CalibrationsOverwatcher(self, calibrations_file)
        self.observer = ObserverOverwatcher(self)
        self.alerts = AlertsOverwatcher(self)
        self.transparency = TransparencyOverwatcher(self)
        self.events = EventsOverwatcher(self)

    async def run(self):
        """Starts the overwatcher tasks."""

        if self.state.running:
            raise GortError("Overwatcher is already running.")

        if not self.gort.is_connected():
            await self.gort.init()

        async with GatheringTaskGroup() as group:
            for module in OverwatcherModule.instances:
                self.log.info(f"Starting overwatcher module {module.name!r}")
                group.create_task(module.run())

        async with GatheringTaskGroup() as group:
            for task in self.tasks:
                group.create_task(task.run())

        self.state.running = True
        await self.notify(
            "Overwatcher is now running.",
            payload={"dry-run": self.state.dry_run},
        )

        if self.state.dry_run:
            self.log.warning("Overatcher is running in dry-mode.")

        return self

    async def shutdown(
        self,
        reason: str = "undefined",
        retry: bool = True,
        park: bool = True,
        disable_overwatcher: bool = False,
    ):
        """Shuts down the observatory."""

        dome_closed = await self.dome.is_closing()
        enabled = self.state.enabled
        observing = self.observer.is_observing

        if dome_closed and not enabled and not observing:
            return

        if not reason.endswith("."):
            reason += "."

        await self.notify(f"Triggering shutdown. Reason: {reason}", level="warning")

        if disable_overwatcher:
            await self.notify("The Overwatcher will be disabled.", level="warning")

        if not self.state.dry_run:
            stop = asyncio.create_task(self.observer.stop_observing(immediate=True))
            shutdown = asyncio.create_task(self.dome.shutdown(retry=retry, park=park))
        else:
            self.log.warning("Dry run enabled. Not shutting down.")
            return

        try:
            await asyncio.gather(stop, shutdown)
        except Exception as err:
            await self.notify(
                f"Error during shutdown: {err!r}",
                level="critical",
                error=err,
            )

        if disable_overwatcher:
            self.state.enabled = False

    async def force_disable(self):
        """Disables the overwatcher."""

        await self.observer.stop_observing(immediate=True)
        await self.calibrations.cancel()

        self.state.enabled = False
        await self.notify("Overwatcher is now disabled.", level="warning")

    async def cancel(self):
        """Cancels the overwatcher tasks."""

        for module in OverwatcherModule.instances:
            self.gort.log.debug(f"Cancelling overwatcher module {module.name!r}")
            await module.cancel()

        for task in self.tasks:
            await task.cancel()

        self.state.running = False
